import asyncio
import contextlib
from os import environ
import logging
from time import sleep, time
from threading import Event, RLock
from typing import Any, Dict, Optional, Union, List, Tuple, cast
from enum import Enum

from math import isclose

from opentrons.drivers.command_builder import CommandBuilder
from serial.serialutil import SerialException  # type: ignore

from opentrons.config.types import RobotConfig
from opentrons.config.robot_configs import current_for_revision
from opentrons.drivers import serial_communication
from opentrons.drivers.types import MoveSplits
from opentrons.drivers.utils import (
    AxisMoveTimestamp, parse_key_values, parse_number, parse_optional_number,
    string_to_hex
)
from opentrons.drivers.rpi_drivers.gpio_simulator import SimulatingGPIOCharDev
from opentrons.drivers.rpi_drivers.dev_types import GPIODriverLike
from opentrons.system import smoothie_update
from . import HOMED_POSITION, Y_BOUND_OVERRIDE


"""
- Driver is responsible for providing an interface for motion control
- Driver is the only system component that knows about GCODES or how smoothie
  communications

- Driver is NOT responsible interpreting the motions in any way
  or knowing anything about what the axes are used for
"""

log = logging.getLogger(__name__)

ERROR_KEYWORD = 'error'
ALARM_KEYWORD = 'alarm'

PLUNGER_BACKLASH_MM = 0.3
LOW_CURRENT_Z_SPEED = 30
CURRENT_CHANGE_DELAY = 0.005
PIPETTE_READ_DELAY = 0.1

Y_SWITCH_BACK_OFF_MM = 28
Y_SWITCH_REVERSE_BACK_OFF_MM = 10
Y_BACKOFF_LOW_CURRENT = 0.8
Y_BACKOFF_SLOW_SPEED = 50
Y_RETRACT_SPEED = 8
Y_RETRACT_DISTANCE = 3

UNSTICK_DISTANCE = 1
UNSTICK_SPEED = 1

DEFAULT_AXES_SPEED = 400

XY_HOMING_SPEED = 80

HOME_SEQUENCE = ['ZABC', 'X', 'Y']
AXES = ''.join(HOME_SEQUENCE)
# Ignore these axis when sending move or home command
DISABLE_AXES = ''

MOVEMENT_ERROR_MARGIN = 1/160  # Largest movement in mm for any step
SEC_PER_MIN = 60

DEFAULT_ACK_TIMEOUT = 5
DEFAULT_EXECUTE_TIMEOUT = 12000
DEFAULT_SMOOTHIE_TIMEOUT = 1
DEFAULT_MOVEMENT_TIMEOUT = 30
SMOOTHIE_BOOT_TIMEOUT = 3
DEFAULT_STABILIZE_DELAY = 0.1

DEFAULT_COMMAND_RETRIES = 3


class GCODE(str, Enum):
    HOME = "G28.2"
    MOVE = "G0"
    DWELL = "G4"
    CURRENT_POSITION = "M114.2"
    LIMIT_SWITCH_STATUS = "M119"
    PROBE = "G38.2"
    ABSOLUTE_COORDS = "G90"
    RELATIVE_COORDS = "G91"
    RESET_FROM_ERROR = "M999"
    PUSH_SPEED = "M120"
    POP_SPEED = "M121"
    SET_SPEED = "G0"
    STEPS_PER_MM = "M92"
    READ_INSTRUMENT_ID = "M369"
    WRITE_INSTRUMENT_ID = "M370"
    READ_INSTRUMENT_MODEL = "M371"
    WRITE_INSTRUMENT_MODEL = "M372"
    SET_MAX_SPEED = "M203.1"
    SET_CURRENT = "M907"
    DISENGAGE_MOTOR = "M18"
    HOMING_STATUS = "G28.6"
    ACCELERATION = "M204"
    WAIT = "M400"
    PIPETTE_RETRACT = "M365.3"
    PIPETTE_DEBOUNCE = "M365.2"
    PIPETTE_MAX_TRAVEL = "M365.1"
    PIPETTE_HOME = "M365.0"
    MICROSTEPPING_B_ENABLE = "M52"
    MICROSTEPPING_B_DISABLE = "M53"
    MICROSTEPPING_C_ENABLE = "M54"
    MICROSTEPPING_C_DISABLE = "M55"
    VERSION = "version"


MICROSTEPPING_GCODES = {
    'B': {
        'ENABLE': GCODE.MICROSTEPPING_B_ENABLE,
        'DISABLE': GCODE.MICROSTEPPING_B_DISABLE,
    },
    'C': {
        'ENABLE': GCODE.MICROSTEPPING_C_ENABLE,
        'DISABLE': GCODE.MICROSTEPPING_C_DISABLE,
    }
}

# Number of digits after the decimal point for coordinates being sent
# to Smoothie
GCODE_ROUNDING_PRECISION = 3

SMOOTHIE_COMMAND_TERMINATOR = '\r\n\r\n'
SMOOTHIE_ACK = 'ok\r\nok\r\n'


class SmoothieError(Exception):
    def __init__(self, ret_code: str = None, command: str = None) -> None:
        self.ret_code = ret_code or ''
        self.command = command
        super().__init__()

    def __repr__(self):
        return f'<SmoothieError: {self.ret_code} from {self.command}>'

    def __str__(self):
        return f'SmoothieError: {self.command} returned {self.ret_code}'


class SmoothieAlarm(Exception):
    def __init__(self, ret_code: str = None, command: str = None) -> None:
        self.ret_code = ret_code
        self.command = command
        super().__init__()

    def __repr__(self):
        return f'<SmoothieAlarm: {self.ret_code} from {self.command}>'

    def __str__(self):
        return f'SmoothieAlarm: {self.command} returned {self.ret_code}'


class TipProbeError(SmoothieAlarm):
    def __init__(self, ret_code: str = None, command: str = None) -> None:
        self.ret_code = ret_code
        self.command = command
        super().__init__(ret_code, command)

    def __repr__(self):
        return f'<TipProbeError: {self.ret_code} from {self.command}'

    def __str__(self):
        return 'Tip probe could not complete: the switch was never touched. '\
            'This may be because there is no tip on the pipette.'


class ParseError(Exception):
    pass


def _parse_position_response(raw_axis_values) -> Dict[str, float]:
    parsed_values = parse_key_values(raw_axis_values)
    if len(parsed_values) < 6:
        msg = f'Unexpected response in _parse_position_response: {raw_axis_values}'
        log.error(msg)
        raise ParseError(msg)

    data = {
        k.title(): parse_number(v, GCODE_ROUNDING_PRECISION)
        for k, v in parsed_values.items()
    }
    return data


def _parse_instrument_data(smoothie_response):
    try:
        items = smoothie_response.split('\n')[0].strip().split(':')
        mount = items[0]
        # data received from Smoothieware is stringified HEX values
        # because of how Smoothieware handles GCODE messages
        data = bytearray.fromhex(items[1])
    except (ValueError, IndexError, TypeError, AttributeError):
        raise ParseError(
            f'Unexpected argument to _parse_instrument_data: {smoothie_response}'
        )
    return {mount: data}


def _byte_array_to_ascii_string(byte_array):
    # remove trailing null characters
    try:
        for c in [b'\x00', b'\xFF']:
            if c in byte_array:
                byte_array = byte_array[:byte_array.index(c)]
        res = byte_array.decode()
    except (ValueError, TypeError, AttributeError):
        log.exception('Unexpected argument to _byte_array_to_ascii_string:')
        raise ParseError(
            f'Unexpected argument to _byte_array_to_ascii_string: {byte_array}'
        )
    return res


def _parse_switch_values(raw_switch_values: str) -> Dict[str, bool]:
    if not raw_switch_values or not isinstance(raw_switch_values, str):
        raise ParseError(
            f'Unexpected argument to _parse_switch_values: {raw_switch_values}'
        )

    # probe has a space after it's ":" for some reason
    if 'Probe: ' in raw_switch_values:
        raw_switch_values = raw_switch_values.replace('Probe: ', 'Probe:')

    parsed_values = parse_key_values(raw_switch_values)
    res = {
        k.title(): bool(parse_optional_number(v, rounding_val=GCODE_ROUNDING_PRECISION))
        for (k, v) in parsed_values.items()
        if any(n in k for n in ['max', 'Probe'])
    }
    # remove the extra "_max" character from each axis key in the dict
    res = {
        key.split('_')[0]: val
        for key, val in res.items()
    }
    if len((list(AXES) + ['Probe']) & res.keys()) != 7:
        raise ParseError(
            f'Unexpected argument to _parse_switch_values: {raw_switch_values}'
        )
    return res


def _parse_homing_status_values(raw_homing_status_values):
    """
    Parse the Smoothieware response to a G28.6 command (homing-status)
    A "1" means it has been homed, and "0" means it has not been homed

    Example response after homing just X axis:
    "X:1 Y:0 Z:0 A:0 B:0 C:0"

    returns: dict
        Key is axis, value is True if the axis needs to be homed
    """
    parsed_values = parse_key_values(raw_homing_status_values)
    res = {
        k.title(): bool(parse_number(v, GCODE_ROUNDING_PRECISION))
        for k, v in parsed_values.items()
    }
    if len(list(AXES) & res.keys()) != 6:
        raise ParseError(
            f'Unexpected argument to '
            f'_parse_homing_status_values: {raw_homing_status_values}'
        )
    return res


def _command_builder() -> CommandBuilder:
    """Create a CommandBuilder"""
    return CommandBuilder(terminator=SMOOTHIE_COMMAND_TERMINATOR)


class SmoothieDriver_3_0_0:
    def __init__(
            self,
            config: RobotConfig,
            gpio_chardev: GPIODriverLike = None,
            handle_locks: bool = True):
        self.run_flag = Event()
        self.run_flag.set()

        self._position = HOMED_POSITION.copy()

        # why do we do this after copying the HOMED_POSITION?
        self._update_position({axis: 0 for axis in AXES})

        self.simulating = True
        self._connection = None
        self._config = config

        self._gpio_chardev = gpio_chardev or SimulatingGPIOCharDev('simulated')

        # Current settings:
        # The amperage of each axis, has been organized into three states:
        # Current-Settings is the amperage each axis was last set to
        # Active-Current-Settings is set when an axis is moving/homing
        # Dwelling-Current-Settings is set when an axis is NOT moving/homing
        self._current_settings: Dict[str, Dict[str, float]] = {
            'now': cast(
                Dict[str, float],
                current_for_revision(
                    config.low_current, self._gpio_chardev.board_rev).copy()),
            'saved': cast(
                Dict[str, float],
                current_for_revision(
                    config.low_current, self._gpio_chardev.board_rev).copy()),
        }
        self._active_current_settings: Dict[str, Dict[str, float]] = {
            'now': cast(
                Dict[str, float],
                current_for_revision(
                    config.high_current, self._gpio_chardev.board_rev).copy()),
            'saved': cast(
                Dict[str, float],
                current_for_revision(
                    config.high_current, self._gpio_chardev.board_rev).copy()),
        }
        self._dwelling_current_settings: Dict[str, Dict[str, float]] = {
            'now': cast(
                Dict[str, float],
                current_for_revision(
                    config.low_current, self._gpio_chardev.board_rev).copy()),
            'saved': cast(
                Dict[str, float],
                current_for_revision(
                    config.low_current, self._gpio_chardev.board_rev).copy()),
        }

        # Active axes are axes that are in use. An axis might be disabled if
        # a motor has had a failure and the robot is operating without that
        # axis until it can be repaired. This will be an unusual circumstance.
        self._active_axes = {ax: False for ax in AXES}

        # Engaged axes are axes that have not been disengaged (GCode M18) since
        # their last "move" or "home" operations. Disengaging an axis stops the
        # power output to the associated motor, primarily for the purpose of
        # reducing heat. When a "disengage" command is sent for an axis, this
        # dict should be updated to False for that axis, and when a "move" or
        # "home" command is sent for an axis, that axis should be updated to
        # True.
        self.engaged_axes = {ax: True for ax in AXES}

        # motor speed settings
        self._max_speed_settings = config.default_max_speed.copy()
        self._saved_max_speed_settings = self._max_speed_settings.copy()
        self._combined_speed = float(DEFAULT_AXES_SPEED)
        self._saved_axes_speed = float(self._combined_speed)
        self._steps_per_mm: Dict[str, float] = {}
        self._acceleration = config.acceleration.copy()
        self._saved_acceleration = config.acceleration.copy()

        # position after homing
        self._homed_position = HOMED_POSITION.copy()
        self.homed_flags: Dict[str, bool] = {}
        self.update_homed_flags(flags={
            'X': False,
            'Y': False,
            'Z': False,
            'A': False,
            'B': False,
            'C': False
        })

        class DummyLock:
            def __enter__(self):
                pass

            def __exit__(self, *args, **kwargs):
                pass

        if handle_locks:
            self._serial_lock: Union[RLock, DummyLock] = RLock()
        else:
            self._serial_lock = DummyLock()
        self._is_hard_halting = Event()
        self._move_split_config: MoveSplits = {}
        #: Cache of currently configured splits from callers
        self._axes_moved_at = AxisMoveTimestamp(AXES)

    @property
    def gpio_chardev(self):
        return self._gpio_chardev

    @gpio_chardev.setter
    def gpio_chardev(self, gpio_chardev):
        self._gpio_chardev = gpio_chardev

    @property
    def homed_position(self) -> Dict[str, float]:
        return self._homed_position.copy()

    @property
    def axis_bounds(self) -> Dict[str, float]:
        bounds = {k: v for k, v in self._homed_position.items()}
        bounds['Y'] = Y_BOUND_OVERRIDE
        return bounds

    def _update_position(self, target):
        self._position.update({
            axis: value
            for axis, value in target.items() if value is not None
        })

    def update_position(self, default=None):
        if default is None:
            default = self._position

        if self.simulating:
            updated_position = self._position.copy()
            updated_position.update(**default)
        else:
            def _recursive_update_position(retries):
                try:
                    position_response = self._send_command(
                        _command_builder().add_gcode(gcode=GCODE.CURRENT_POSITION)
                    )
                    return _parse_position_response(position_response)
                except ParseError as e:
                    retries -= 1
                    if retries <= 0:
                        raise e
                    if not self.simulating:
                        sleep(DEFAULT_STABILIZE_DELAY)
                    return _recursive_update_position(retries)

            updated_position = _recursive_update_position(
                DEFAULT_COMMAND_RETRIES)

        self._update_position(updated_position)

    def configure_splits_for(self, config: MoveSplits):
        """ Configure the driver to automatically split moves on a given
        axis that execute (including pauses) after a specified amount of
        time. The move created will adhere to the split config.

        To remove the setting, set None for the specified axis.

        Only pipette axes may be specified for splitting
        """
        assert all((ax.lower() in 'bc' for ax in config.keys())),\
            'splits may only be configured for plunger axes'
        self._move_split_config.update(config)
        log.info(f"Updated move split config with {config}")
        self._axes_moved_at.reset_moved(config.keys())

    def read_pipette_id(self, mount) -> Optional[str]:
        """
        Reads in an attached pipette's ID
        The ID is unique to this pipette, and is a string of unknown length

        :param mount: string with value 'left' or 'right'
        :return id string, or None
        """
        res: Optional[str] = None
        if self.simulating:
            res = '1234567890'
        else:
            try:
                res = self._read_from_pipette(GCODE.READ_INSTRUMENT_ID, mount)
            except UnicodeDecodeError:
                log.exception("Failed to decode pipette ID string:")
                res = None
        return res

    def read_pipette_model(self, mount) -> Optional[str]:
        """
        Reads an attached pipette's MODEL
        The MODEL is a unique string for this model of pipette

        :param mount: string with value 'left' or 'right'
        :return model string, or None
        """
        if self.simulating:
            res = None
        else:
            res = self._read_from_pipette(
                GCODE.READ_INSTRUMENT_MODEL, mount)
            if res and '_v' not in res:
                # Backward compatibility for pipettes programmed with model
                # strings that did not include the _v# designation
                res = res + '_v1'
            elif res and '_v13' in res:
                # Backward compatibility for pipettes programmed with model
                # strings that did not include the "." to seperate version
                # major and minor values
                res = res.replace('_v13', '_v1.3')

        return res

    def write_pipette_id(self, mount: str, data_string: str):
        """
        Writes to an attached pipette's ID memory location
        The ID is unique to this pipette, and is a string of unknown length

        NOTE: To enable write-access to the pipette, it's button must be held

        mount:
            String (str) with value 'left' or 'right'
        data_string:
            String (str) that is of unknown length, and should be unique to
            this one pipette
        """
        self._write_to_pipette(
            GCODE.WRITE_INSTRUMENT_ID, mount, data_string)

    def write_pipette_model(self, mount: str, data_string: str):
        """
        Writes to an attached pipette's MODEL memory location
        The MODEL is a unique string for this model of pipette

        NOTE: To enable write-access to the pipette, it's button must be held

        mount:
            String (str) with value 'left' or 'right'
        data_string:
            String (str) that is unique to this model of pipette
        """
        self._write_to_pipette(GCODE.WRITE_INSTRUMENT_MODEL, mount, data_string)

    def update_pipette_config(
            self, axis: str, data: Dict[str, float])\
            -> Dict[str, Dict[str, float]]:
        """
        Updates the following configs for a given pipette mount based on
        the detected pipette type:
        - homing positions M365.0
        - Max Travel M365.1
        - endstop debounce M365.2 (NOT for zprobe debounce)
        - retract from endstop distance M365.3

        Returns the data as the value of a dict with the axis as a key.
        For instance, calling update_pipette_config('B', {'retract': 2})
        would return (if successful) {'B': {'retract': 2}}
        """
        if self.simulating:
            return {axis: data}

        gcodes = {
            'retract': GCODE.PIPETTE_RETRACT,
            'debounce': GCODE.PIPETTE_DEBOUNCE,
            'max_travel': GCODE.PIPETTE_MAX_TRAVEL,
            'home': GCODE.PIPETTE_HOME
        }

        res_msg: Dict[str, Dict[str, float]] = {axis: {}}

        for key, value in data.items():
            cmd = _command_builder().add_gcode(
                gcode=gcodes[key]
            )
            if key == 'debounce':
                # debounce variable for all axes, so do not specify an axis
                cmd.add_float(prefix='O', value=value, precision=None)
            else:
                cmd.add_float(prefix=axis, value=value, precision=None)
            res = self._send_command(cmd)
            if res is None:
                raise ValueError(
                    f'{key} was not updated to {value} on {axis} axis')
            res_msg[axis][key] = value

        return res_msg

    # FIXME (JG 9/28/17): Should have a more thought out
    # way of simulating vs really running
    def connect(self, port: str = None):
        if environ.get('ENABLE_VIRTUAL_SMOOTHIE', '').lower() == 'true':
            self.simulating = True
            return
        self.disconnect()
        self._connect_to_port(port)
        self._setup()

    def disconnect(self):
        if self.is_connected():
            self._connection.close()  # type: ignore
        self._connection = None
        self.simulating = True

    def is_connected(self) -> bool:
        if not self._connection:
            return False
        return self._connection.is_open

    def _connect_to_port(self, port: str = None):
        try:
            smoothie_id = environ.get('OT_SMOOTHIE_ID', 'AMA')
            self._connection = serial_communication.connect(
                device_name=smoothie_id,
                port=port,
                baudrate=self._config.serial_speed
            )
            self.simulating = False
        except SerialException:
            # if another process is using the port, pyserial raises an
            # exception that describes a "readiness to read" which is confusing
            error_msg = 'Unable to access UART port to Smoothie. This is '
            error_msg += 'because another process is currently using it, or '
            error_msg += 'the UART port is disabled on this device (OS)'
            raise SerialException(error_msg)

    @property
    def port(self) -> Optional[str]:
        if not self._connection:
            return None
        return self._connection.port

    def get_fw_version(self) -> str:
        """
        Queries Smoothieware for it's build version, and returns
        the parsed response.

        returns: str
            Current version of attached Smoothi-driver. Versions are derived
            from git branch-hash (eg: edge-66ec883NOMSD)

        Example Smoothieware response:

        Build version: edge-66ec883NOMSD, Build date: Jan 28 2018 15:26:57, MCU: LPC1769, System Clock: 120MHz  # NOQA
          CNC Build   NOMSD Build
        6 axis
        """
        version = 'Virtual Smoothie'
        if not self.simulating:
            version = self._send_command(_command_builder().add_gcode(
                gcode=GCODE.VERSION))
            version = version.split(',')[0].split(':')[-1].strip()
            version = version.replace('NOMSD', '')
        return version

    @property
    def position(self) -> Dict[str, float]:
        """
        Instead of sending M114.2 we are storing target values in
        self._position since movement and home commands are blocking and
        assumed to go the correct place.

        Cases where Smoothie would not be in the correct place (such as if a
        belt slips) would not be corrected by getting position with M114.2
        because Smoothie would also not be aware of slippage.
        """
        return {k.upper(): v for k, v in self._position.items()}

    @property
    def switch_state(self) -> Dict[str, bool]:
        """Returns the state of all SmoothieBoard limit switches"""
        res = self._send_command(_command_builder().add_gcode(
            gcode=GCODE.LIMIT_SWITCH_STATUS
        ))
        return _parse_switch_values(res)

    def update_homed_flags(
            self, flags: Dict[str, bool] = None):
        """
        Returns Smoothieware's current homing-status, which is a dictionary
        of boolean values for each axis (XYZABC). If an axis is False, then it
        still needs to be homed, and it's coordinate cannot be trusted.
        Smoothieware sets it's internal homing flags for all axes to False when
        it has yet to home since booting/restarting, or an endstop/homing error
        """
        if flags and isinstance(flags, dict):
            self.homed_flags.update(flags)

        elif self.simulating:
            self.homed_flags.update({ax: False for ax in AXES})

        elif self.is_connected():

            def _recursive_update_homed_flags(retries: int):
                try:
                    res = self._send_command(
                        _command_builder().add_gcode(gcode=GCODE.HOMING_STATUS))
                    flags = _parse_homing_status_values(res)
                    self.homed_flags.update(flags)
                except ParseError as e:
                    retries -= 1
                    if retries <= 0:
                        raise e
                    if not self.simulating:
                        sleep(DEFAULT_STABILIZE_DELAY)
                    return _recursive_update_homed_flags(retries)

            _recursive_update_homed_flags(DEFAULT_COMMAND_RETRIES)

    @property
    def current(self) -> Dict[str, float]:
        return self._current_settings['now']

    @property
    def speed(self):
        pass

    @property
    def steps_per_mm(self) -> Dict[str, float]:
        return self._steps_per_mm

    @contextlib.contextmanager
    def restore_speed(self, value: Union[float, str]):
        self.set_speed(value, update=False)
        try:
            yield
        finally:
            self.set_speed(self._combined_speed)

    def _build_speed_command(self, speed: float) -> CommandBuilder:
        return _command_builder().add_gcode(
            gcode=GCODE.SET_SPEED
        ).add_int(prefix="F", value=int(float(speed) * SEC_PER_MIN))

    def set_speed(self, value: Union[float, str], update: bool = True):
        """ set total axes movement speed in mm/second"""
        if update:
            self._combined_speed = float(value)
        command = self._build_speed_command(float(value))
        log.debug(f"set_speed: {command}")
        self._send_command(command)

    def push_speed(self):
        self._saved_axes_speed = float(self._combined_speed)

    def pop_speed(self):
        self.set_speed(self._saved_axes_speed)

    @contextlib.contextmanager
    def restore_axis_max_speed(self, new_max_speeds: Dict[str, float]):
        self.set_axis_max_speed(new_max_speeds, update=False)
        try:
            yield
        finally:
            self.set_axis_max_speed(self._max_speed_settings)  # type: ignore

    def set_axis_max_speed(
            self, settings: Dict[str, float], update: bool = True):
        """
        Sets the maximum speed (mm/sec) that a given axis will move

        settings
            Dict with axes as valies (e.g.: 'X', 'Y', 'Z', 'A', 'B', or 'C')
            and floating point number for millimeters per second (mm/sec)
        update
            bool, True to save the settings for future use
        """
        if update:
            self._max_speed_settings.update(settings)  # type: ignore

        command = _command_builder().add_gcode(gcode=GCODE.SET_MAX_SPEED)
        for axis, value in sorted(settings.items()):
            command = command.add_float(
                prefix=axis, value=value, precision=None
            )

        log.debug(f"set_axis_max_speed: {command}")
        self._send_command(command)

    def push_axis_max_speed(self):
        self._saved_max_speed_settings = self._max_speed_settings.copy()

    def pop_axis_max_speed(self):
        self.set_axis_max_speed(self._saved_max_speed_settings)  # type: ignore

    def set_acceleration(self, settings: Dict[str, float]):
        """
        Sets the acceleration (mm/sec^2) that a given axis will move

        settings
            Dict with axes as valies (e.g.: 'X', 'Y', 'Z', 'A', 'B', or 'C')
            and floating point number for mm-per-second-squared (mm/sec^2)
        """
        self._acceleration.update(settings)

        command = _command_builder().add_gcode(
            gcode=GCODE.ACCELERATION
        ).add_int(
            prefix="S", value=10000
        )
        for axis, value in sorted(settings.items()):
            command.add_float(prefix=axis, value=value, precision=None)

        log.debug(f"set_acceleration: {command}")
        self._send_command(command)

    def push_acceleration(self):
        self._saved_acceleration = self._acceleration.copy()

    def pop_acceleration(self):
        self.set_acceleration(self._saved_acceleration)

    def set_active_current(self, settings: Dict[str, float]):
        """
        Sets the amperage of each motor for when it is activated by driver.
        Values are initialized from the `robot_config.high_current` values,
        and can then be changed through this method by other parts of the API.

        For example, `Pipette` setting the active-current of it's pipette,
        depending on what model pipette it is, and what action it is performing

        settings
            Dict with axes as valies (e.g.: 'X', 'Y', 'Z', 'A', 'B', or 'C')
            and floating point number for current (generally between 0.1 and 2)
        """
        self._active_current_settings['now'].update(settings)

        # if an axis specified in the `settings` is currently active,
        # reset it's current to the new active-current value
        active_axes_to_update = {
            axis: amperage
            for axis, amperage in self._active_current_settings['now'].items()
            if self._active_axes.get(axis) is True
            if self.current[axis] != amperage
        }
        if active_axes_to_update:
            self._save_current(active_axes_to_update, axes_active=True)

    def push_active_current(self):
        self._active_current_settings['saved'].update(
            self._active_current_settings['now'])

    def pop_active_current(self):
        self.set_active_current(self._active_current_settings['saved'])

    def set_dwelling_current(self, settings: Dict[str, float]):
        """
        Sets the amperage of each motor for when it is dwelling.
        Values are initialized from the `robot_config.log_current` values,
        and can then be changed through this method by other parts of the API.

        For example, `Pipette` setting the dwelling-current of it's pipette,
        depending on what model pipette it is.

        settings
            Dict with axes as valies (e.g.: 'X', 'Y', 'Z', 'A', 'B', or 'C')
            and floating point number for current (generally between 0.1 and 2)
        """
        self._dwelling_current_settings['now'].update(settings)

        # if an axis specified in the `settings` is currently dwelling,
        # reset it's current to the new dwelling-current value
        dwelling_axes_to_update = {
            axis: amps
            for axis, amps in self._dwelling_current_settings['now'].items()
            if self._active_axes.get(axis) is False
            if self.current[axis] != amps
        }
        if dwelling_axes_to_update:
            self._save_current(dwelling_axes_to_update, axes_active=False)

    def push_dwelling_current(self):
        self._dwelling_current_settings['saved'].update(
            self._dwelling_current_settings['now'])

    def pop_dwelling_current(self):
        self.set_dwelling_current(self._dwelling_current_settings['saved'])

    def _save_current(
            self, settings: Dict[str, float], axes_active: bool = True):
        """
        Sets the current in Amperes (A) by axis. Currents are limited to be
        between 0.0-2.0 amps per axis motor.

        Note: this method does not send gcode commands, but instead stores the
        desired current setting. A seperate call to _generate_current_command()
        will return a gcode command that can be used to set Smoothie's current

        settings
            Dict with axes as valies (e.g.: 'X', 'Y', 'Z', 'A', 'B', or 'C')
            and floating point number for current (generally between 0.1 and 2)
        """
        self._active_axes.update({
            ax: axes_active
            for ax in settings.keys()
        })
        self._current_settings['now'].update(settings)
        log.debug(f"_save_current: {self.current}")

    def _set_saved_current(self):
        """
        Sends the driver's current settings to the serial port as gcode. Call
        this method to set the axis-current state on the actual Smoothie
        motor-driver.
        """
        self._send_command(self._generate_current_command())

    def _generate_current_command(self) -> CommandBuilder:
        """
        Returns a constructed GCode string that contains this driver's
        axis-current settings, plus a small delay to wait for those settings
        to take effect.
        """
        command = _command_builder().add_gcode(gcode=GCODE.SET_CURRENT)
        for axis, value in sorted(self.current.items()):
            command.add_float(prefix=axis, value=value, precision=None)

        command.add_gcode(
            gcode=GCODE.DWELL
        ).add_float(
            prefix="P", value=CURRENT_CHANGE_DELAY, precision=None
        )
        log.debug(f"_generate_current_command: {command}")
        return command

    def disengage_axis(self, axes: str):
        """
        Disable the stepper-motor-driver's 36v output to motor
        This is a safe GCODE to send to Smoothieware, as it will automatically
        re-engage the motor if it receives a home or move command

        axes
            String containing the axes to be disengaged
            (e.g.: 'XY' or 'ZA' or 'XYZABC')
        """
        axes = ''.join(set(axes.upper()) & set(AXES))
        if axes:
            log.debug(f"disengage_axis: {axes}")
            self._send_command(
                _command_builder().add_gcode(
                    gcode=GCODE.DISENGAGE_MOTOR
                ).add_element(element=axes))
            for axis in axes:
                self.engaged_axes[axis] = False

    def dwell_axes(self, axes: str):
        """
        Sets motors to low current, for when they are not moving.

        Dwell for XYZA axes is only called after HOMING
        Dwell for BC axes is called after both HOMING and MOVING

        axes:
            String containing the axes to set to low current (eg: 'XYZABC')
        """
        axes = ''.join(set(axes) & set(AXES) - set(DISABLE_AXES))
        dwelling_currents = {
            ax: self._dwelling_current_settings['now'][ax]
            for ax in axes
            if self._active_axes[ax] is True
        }
        if dwelling_currents:
            self._save_current(dwelling_currents, axes_active=False)

    def activate_axes(self, axes: str):
        """
        Sets motors to a high current, for when they are moving
        and/or must hold position

        Activating XYZABC axes before both HOMING and MOVING

        axes:
            String containing the axes to set to high current (eg: 'XYZABC')
        """
        axes = ''.join(set(axes) & set(AXES) - set(DISABLE_AXES))
        active_currents = {
            ax: self._active_current_settings['now'][ax]
            for ax in axes
            if self._active_axes[ax] is False
        }
        if active_currents:
            self._save_current(active_currents, axes_active=True)

    # ----------- Private functions --------------- #

    def _wait_for_ack(self):
        """
        In the case where smoothieware has just been reset, we want to
        ignore all the garbage it spits out

        This methods writes a sequence of newline characters, which will
        guarantee Smoothieware responds with 'ok\r\nok\r\n' within 3 seconds
        """
        self._send_command(_command_builder(), timeout=SMOOTHIE_BOOT_TIMEOUT)

    def _reset_from_error(self):
        # smoothieware will ignore new messages for a short time
        # after it has entered an error state, so sleep for some milliseconds
        if not self.simulating:
            sleep(DEFAULT_STABILIZE_DELAY)
        log.debug("reset_from_error")
        self._send_command(
            _command_builder().add_gcode(gcode=GCODE.RESET_FROM_ERROR)
        )
        self.update_homed_flags()

    # Potential place for command optimization (buffering, flushing, etc)
    def _send_command(
            self,
            command: CommandBuilder,
            timeout: float = DEFAULT_EXECUTE_TIMEOUT,
            suppress_error_msg: bool = False,
            ack_timeout: float = DEFAULT_ACK_TIMEOUT,
            suppress_home_after_error: bool = False):
        """
        Submit a GCODE command to the robot, followed by M400 to block until
        done. This method also ensures that any command on the B or C axis
        (the axis for plunger control) do current ramp-up and ramp-down, so
        that plunger motors rest at a low current to prevent burn-out.

        In the case of a limit-switch alarm during any command other than home,
        the robot should home the axis from the alarm and then raise a
        SmoothieError. The robot should *not* recover and continue to run the
        protocol, as this could result in unpredictable handling of liquids.
        When a SmoothieError is raised, the user should inspect the physical
        configuration of the robot and the protocol and determine why the limit
        switch was hit unexpectedly. This is usually due to an undetected
        collision in a previous move command.

        SmoothieErrors are also raised when a command is sent to a pipette that
        is not present, such as when identifying which pipettes are on a robot.
        In this case, the message should not be logged, so the caller of this
        function should specify `supress_error_msg=True`.

        :param command: the GCODE to submit to the robot
        :param timeout: the time to wait for the smoothie to execute the
            command (after an m400). this should be long enough to allow the
            command to execute. If this is None, the timeout will be infinite.
            This is almost certainly not what you want.
        :param suppress_error_msg: flag for indicating that smoothie errors
            should not be logged
        :param ack_timeout: The time to wait for the smoothie to ack a
            command. For commands that queue (like move) or are short (like
            pipette interrogation) this should be a small number, and is the
            default. For commands the smoothie only acks after execution,
            like home, it should be long enough to allow the command to
            complete in the worst case. If this is None, the timeout will
            be infinite. This is almost certainly not what you want.
        """
        if self.simulating:
            return
        try:
            with self._serial_lock:
                return self._send_command_unsynchronized(command,
                                                         ack_timeout,
                                                         timeout)
        except SmoothieError as se:
            # XXX: This is a reentrancy error because another command could
            # swoop in here. We're already resetting though and errors (should
            # be) rare so it's probably fine, but the actual solution to this
            # is locking at a higher level like in APIv2.
            self._reset_from_error()
            error_axis = se.ret_code.strip()[-1]
            if not suppress_error_msg:
                log.warning(
                        f"alarm/error: command={command}, resp={se.ret_code}")
            if (GCODE.MOVE in command or GCODE.PROBE in command)\
               and not suppress_home_after_error:
                if error_axis not in 'XYZABC':
                    error_axis = AXES
                log.info("Homing after alarm/error")
                self.home(error_axis)
            raise SmoothieError(se.ret_code, str(command))

    def _send_command_unsynchronized(self,
                                     command: CommandBuilder,
                                     ack_timeout: float,
                                     execute_timeout: float):
        command_str = command.build()
        cmd_ret = self._write_with_retries(
            command_str,
            ack_timeout, DEFAULT_COMMAND_RETRIES)
        cmd_ret = self._remove_unwanted_characters(command_str, cmd_ret)
        self._handle_return(cmd_ret)
        wait_ret = serial_communication.write_and_return(
            _command_builder().add_gcode(gcode=GCODE.WAIT).build(),
            SMOOTHIE_ACK, self._connection, timeout=execute_timeout,
            tag='smoothie')
        wait_ret = self._remove_unwanted_characters(
            _command_builder().add_gcode(gcode=GCODE.WAIT).build(),
            wait_ret)
        self._handle_return(wait_ret)
        return cmd_ret.strip()

    def _handle_return(self, ret_code: str):
        """ Check the return string from smoothie for an error condition.

        Usually raises a SmoothieError, which can be handled by the error
        handling in write_with_retries. However, if the hard halt line has
        been set, we need to catch that halt and _not_ handle it, since it
        is used for things like cancelling protocols and needs to be
        handled elsewhere. In that case, we raise SmoothieAlarm, which isn't
        (and shouldn't be) handled by the normal error handling.
        """
        is_alarm = ALARM_KEYWORD in ret_code.lower()
        is_error = ERROR_KEYWORD in ret_code.lower()
        if self._is_hard_halting.is_set():
            # This is the alarm from setting the hard halt
            if is_alarm:
                self._is_hard_halting.clear()
                raise SmoothieAlarm(ret_code)
            elif is_error:
                # this would be a race condition
                raise SmoothieError(ret_code)
        else:
            if is_alarm or is_error:
                # info-level logging for errors of form "no L instrument found"
                if 'instrument found' in ret_code.lower():
                    log.info(f"smoothie: {ret_code}")
                    raise SmoothieError(ret_code)

                # the two errors below happen when we're recovering from a hard
                # halt. in that case, some try/finallys above us may send
                # further commands. smoothie responds to those commands with
                # errors like these. if we raise exceptions here, they
                # overwrite the original exception and we don't properly
                #  handle it. This hack to get around this is really bad!
                if 'alarm lock' not in ret_code.lower()\
                   and 'after halt you should home' not in ret_code.lower():
                    log.error(f"alarm/error outside hard halt: {ret_code}")
                    raise SmoothieError(ret_code)

    def _remove_unwanted_characters(self, command: str, response: str) -> str:
        # smoothieware can enter a weird state, where it repeats back
        # the sent command at the beginning of its response.
        # Check for this echo, and strips the command from the response
        def _is_token_command(_s: str) -> bool:
            """check if token is a command"""
            # A single letter token cannot be assumed to be a command.
            # For example: "M369 L" response is "L:2132121212".
            return len(_s) > 1
        # Split at spaces.
        tokens = (c.strip() for c in command.strip().split(' '))
        # A list of commands to remove from response. Including the entire
        # command.
        remove_from_response = [command] + [
            c for c in tokens if _is_token_command(c)
        ]

        # also removing any inadvertent newline/return characters
        # this is ok because all data we need from Smoothie is returned on
        # the first line in the response
        remove_from_response += ['\r', '\n']
        modified_response = str(response)

        for cmd in remove_from_response:
            modified_response = modified_response.replace(cmd, '')

        if modified_response != response:
            log.debug(f'Removed characters from response: {response}')
            log.debug(f'Newly formatted response: {modified_response}')

        return modified_response

    def _write_with_retries(self, cmd: str, timeout: float, retries: int):
        for attempt in range(retries):
            try:
                ret = serial_communication.write_and_return(
                    cmd,
                    SMOOTHIE_ACK,
                    self._connection,
                    timeout=timeout,
                    tag='smoothie')
                if attempt != 0:
                    log.warning(
                        f"required {attempt} retries for {cmd.strip()}")
                return ret
            except serial_communication.SerialNoResponse:
                if not self.simulating:
                    sleep(DEFAULT_STABILIZE_DELAY)
                if self._connection:
                    self._connection.close()
                    self._connection.open()
        raise serial_communication.SerialNoResponse()

    def _home_x(self):
        log.debug("_home_x")
        # move the gantry forward on Y axis with low power
        self._save_current({'Y': Y_BACKOFF_LOW_CURRENT})
        self.push_axis_max_speed()
        self.set_axis_max_speed({'Y': Y_BACKOFF_SLOW_SPEED})

        # move away from the Y endstop switch, then backward half that distance
        relative_retract_command = _command_builder().add_gcode(
            # set to relative coordinate system
            gcode=GCODE.RELATIVE_COORDS
        ).add_gcode(
            gcode=GCODE.MOVE
        ).add_int(
            # move towards front of machine
            prefix="Y", value=int(-Y_SWITCH_BACK_OFF_MM)
        ).add_gcode(
            gcode=GCODE.MOVE
        ).add_int(
            # move towards back of machine
            prefix="Y", value=int(Y_SWITCH_REVERSE_BACK_OFF_MM)
        ).add_gcode(
            # set back to abs coordinate system
            gcode=GCODE.ABSOLUTE_COORDS
        )

        command = self._generate_current_command().add_builder(
            builder=relative_retract_command
        )
        self._send_command(command)
        self.dwell_axes('Y')

        # time it is safe to home the X axis
        try:
            # override firmware's default XY homing speed, to avoid resonance
            self.set_axis_max_speed({'X': XY_HOMING_SPEED})
            self.activate_axes('X')
            command = self._generate_current_command().add_gcode(
                gcode=GCODE.HOME
            ).add_element("X")
            # home commands are acked after execution rather than queueing, so
            # we want a long ack timeout and a short execution timeout
            home_timeout = (HOMED_POSITION['X'] / XY_HOMING_SPEED) * 2
            self._send_command(command, ack_timeout=home_timeout,
                               timeout=5)
            self.update_homed_flags(flags={'X': True})
        finally:
            self.pop_axis_max_speed()
            self.dwell_axes('X')
            self._set_saved_current()

    def _home_y(self):
        log.debug("_home_y")
        # override firmware's default XY homing speed, to avoid resonance
        self.push_axis_max_speed()
        self.set_axis_max_speed({'Y': XY_HOMING_SPEED})

        self.activate_axes('Y')
        # home the Y at normal speed (fast)
        command = self._generate_current_command().add_gcode(
            gcode=GCODE.HOME
        ).add_element("Y")
        fast_home_timeout = (HOMED_POSITION['Y'] / XY_HOMING_SPEED) * 2
        # home commands are executed before ack, set a long ack timeout
        self._send_command(command, ack_timeout=fast_home_timeout,
                           timeout=5)

        # slow the maximum allowed speed on Y axis
        self.set_axis_max_speed({'Y': Y_RETRACT_SPEED})

        # retract, then home, then retract again
        relative_retract_command = _command_builder().add_gcode(
            # set to relative coordinate system
            gcode=GCODE.RELATIVE_COORDS
        ).add_gcode(
            gcode=GCODE.MOVE
        ).add_int(
            # move 3 millimeters away from switch
            prefix="Y", value=-Y_RETRACT_DISTANCE
        ).add_gcode(
            # set back to abs coordinate system
            gcode=GCODE.ABSOLUTE_COORDS
        )
        try:
            self._send_command(relative_retract_command)
            # home commands are executed before ack, use a long ack timeout
            slow_timeout = (Y_RETRACT_DISTANCE / Y_RETRACT_SPEED) * 2
            self._send_command(
                _command_builder().add_gcode(gcode=GCODE.HOME).add_element("Y"),
                ack_timeout=slow_timeout,
                timeout=5)
            self.update_homed_flags(flags={'Y': True})
            self._send_command(
                relative_retract_command)
        finally:
            self.pop_axis_max_speed()  # bring max speeds back to normal
            self.dwell_axes('Y')
            self._set_saved_current()

    def _setup(self):
        log.debug("_setup")
        try:
            self._wait_for_ack()
        except serial_communication.SerialNoResponse:
            # incase motor-driver is stuck in bootloader and unresponsive,
            # use gpio to reset into a ktimen state
            log.debug("wait for ack failed, resetting")
            self._smoothie_reset()
        log.debug("wait for ack done")
        self._reset_from_error()
        log.debug("_reset")
        self.update_steps_per_mm(self._config.gantry_steps_per_mm)
        self.update_steps_per_mm({
            ax: self._config.default_pipette_configs['stepsPerMM']
            for ax in 'BC'})
        log.debug("sent steps")
        self._send_command(
            _command_builder().add_gcode(gcode=GCODE.ABSOLUTE_COORDS)
        )
        log.debug("sent abs")
        self._save_current(self.current, axes_active=False)
        log.debug("sent current")
        self.update_position(default=self.homed_position)
        self.pop_axis_max_speed()
        self.pop_speed()
        self.pop_acceleration()
        log.debug("setup done")

    def _build_steps_per_mm(self, data: Dict[str, float]) -> CommandBuilder:
        """ Build the set steps/mm command string without sending """
        command = _command_builder()

        if not data:
            return command

        command.add_gcode(
            gcode=GCODE.STEPS_PER_MM
        )
        for axis, value in data.items():
            command.add_float(prefix=axis, value=value, precision=None)
        return command

    def update_steps_per_mm(self, data: Union[Dict[str, float], str]):
        # Using M92, update steps per mm for a given axis
        if self.simulating:
            if isinstance(data, dict):
                self.steps_per_mm.update(data)
            return

        if isinstance(data, str):
            # Unfortunately update server calls driver._setup() before the
            # update can correctly load the robot_config change on disk.
            # Need to account for old command format to avoid this issue.
            self._send_command(_command_builder().add_gcode(data))
        else:
            self.steps_per_mm.update(data)
            cmd = self._build_steps_per_mm(data)
            self._send_command(cmd)

    def _read_from_pipette(self, gcode: str, mount: str) -> Optional[str]:
        """
        Read from an attached pipette's internal memory. The gcode used
        determines which portion of memory is read and returned.

        All motors must be disengaged to consistently read over I2C lines

        gcode:
            String (str) containing a GCode
            either 'READ_INSTRUMENT_ID' or 'READ_INSTRUMENT_MODEL'
        mount:
            String (str) with value 'left' or 'right'
        """
        allowed_mounts = {'left': 'L', 'right': 'R'}
        allowed_mount = allowed_mounts.get(mount)
        if not allowed_mount:
            raise ValueError(f'Unexpected mount: {mount}')
        try:
            # EMI interference from both plunger motors has been found to
            # prevent the I2C lines from communicating between Smoothieware and
            # pipette's onboard EEPROM. To avoid, turn off both plunger motors
            self.disengage_axis('ZABC')
            self.delay(PIPETTE_READ_DELAY)
            # request from Smoothieware the information from that pipette
            res = self._send_command(
                _command_builder().add_gcode(gcode=gcode).add_element(allowed_mount),
                suppress_error_msg=True)
            if res:
                res = _parse_instrument_data(res)
                assert allowed_mount in res
                # data is read/written as strings of HEX characters
                # to avoid firmware weirdness in how it parses GCode arguments
                return _byte_array_to_ascii_string(res[allowed_mount])
        except (ParseError, AssertionError, SmoothieError):
            pass
        return None

    def _write_to_pipette(self, gcode: str, mount: str, data_string: str):
        """
        Write to an attached pipette's internal memory. The gcode used
        determines which portion of memory is written to.

        NOTE: To enable write-access to the pipette, it's button must be held

        gcode:
            String (str) containing a GCode
            either 'WRITE_INSTRUMENT_ID' or 'WRITE_INSTRUMENT_MODEL'
        mount:
            String (str) with value 'left' or 'right'
        data_string:
            String (str) that is of unkown length
        """
        allowed_mounts = {'left': 'L', 'right': 'R'}
        allowed_mount = allowed_mounts.get(mount)
        if not allowed_mount:
            raise ValueError(f'Unexpected mount: {mount}')
        if not isinstance(data_string, str):
            raise ValueError(
                'Expected {0}, not {1}'.format(str, type(data_string)))
        # EMI interference from both plunger motors has been found to
        # prevent the I2C lines from communicating between Smoothieware and
        # pipette's onboard EEPROM. To avoid, turn off both plunger motors
        self.disengage_axis('BC')
        self.delay(CURRENT_CHANGE_DELAY)
        # data is read/written as strings of HEX characters
        # to avoid firmware weirdness in how it parses GCode arguments
        byte_string = string_to_hex(data_string)
        command = _command_builder().add_gcode(
            gcode=gcode
        ).add_element(
            element=allowed_mount
        ).add_element(
            element=byte_string
        )
        log.debug(f"_write_to_pipette: {command}")
        self._send_command(command)

    # ----------- END Private functions ----------- #

    # ----------- Public interface ---------------- #
    def move(self, target: Dict[str, float], home_flagged_axes: bool = False,  # noqa: C901, E501
             speed: float = None):
        """
        Move to the `target` Smoothieware coordinate, along any of the size
        axes, XYZABC.

        :param target: dict setting the coordinate that Smoothieware will be
            at when `move()` returns. `target` keys are the axis in
            upper-case, and the values are the coordinate in mm (float)
        :param home_flagged_axes: boolean (default=False)
            If set to `True`, each axis included within the target coordinate
            may be homed before moving, determined by Smoothieware's internal
            homing-status flags (`True` means it has already homed). All axes'
            flags are set to `False` by Smoothieware under three conditions:
            1) Smoothieware boots or resets, 2) if a HALT gcode or signal
            is sent, or 3) a homing/limitswitch error occured.
        :param speed: Optional speed for the move. If not specified, set to the
            current cached _combined_speed. To avoid conflict with callers that
            expect the smoothie's speed setting to always be combined_speed,
            the smoothie is set back to this state after every move


        If the current move split config indicates that the move should be
        broken up, the driver will do so. If the new position requires a
        change in position of an axis with a split configuration, it may be
        split into multiple moves such that the axis will move a maximum of the
        specified split distance at the specified current and speed. If the
        axis would move less than the split distance, it will move the
        entire distance at the specified current and speed.

        This command respects the run flag and will wait until it is set.

        The function may issue up to 3 moves:
        - if move splitting is required, the split move
        - the actual move, plus a bit extra to give room to preload backlash
        - if we preload backlash we then issue a third move to preload backlash
        """
        self.run_flag.wait()

        def valid_movement(axis: str, coord: float) -> bool:
            """ True if the axis is not disabled and the coord is different
            from the current position cache
            """
            return not (
                (axis in DISABLE_AXES) or
                isclose(coord, self.position[axis],
                        rel_tol=1e-05, abs_tol=1e-08)
            )

        def only_moving(move_target: Dict[str, float]) -> Dict[str, float]:
            """ Filter a target dict to have only those axes which have valid
            movements"""
            return {ax: coord for ax, coord in move_target.items()
                    if valid_movement(ax, coord)}

        def create_coords_list(coords_dict: Dict[str, float]) -> CommandBuilder:
            """ Build the gcode string for a move """
            cmd = _command_builder()
            for axis, coords in sorted(coords_dict.items()):
                if valid_movement(axis, coords):
                    cmd.add_float(prefix=axis, value=coords,
                                  precision=GCODE_ROUNDING_PRECISION)
            return cmd

        moving_target = only_moving(target)
        if not moving_target:
            log.info(
                f"No axes move in {target} from position {self.position}")
            return

        backlash_target = {
            axis: value + PLUNGER_BACKLASH_MM
            for axis, value in target.items()
            if axis in 'BC' and self.position[axis] < value
        }

        # whatever else we do to our motion target, if nothing moves in the
        # input we will not command it to move
        non_moving_axes = [ax for ax in AXES if ax not in moving_target.keys()]

        # cache which axes move because we might take them out of moving target
        moving_axes = list(moving_target.keys())

        def build_split(
                here: float, dest: float, split_distance: float) -> float:
            """ Return the destination for the split move """
            if dest < here:
                return max(dest, here-split_distance)
            else:
                return min(dest, here+split_distance)

        since_moved = self._axes_moved_at.time_since_moved()
        # generate the split moves if necessary
        split_target = {
            ax: build_split(
                self.position[ax],
                backlash_target.get(ax, moving_target[ax]),
                split.split_distance)
            for ax, split in self._move_split_config.items()
            # a split is only necessary if:
            # - the axis is moving
            if (ax in moving_target)
            # - we have a split configuration
            and split
            # - it's been long enough since the last time it moved
            and ((since_moved[ax] is None)
                 or (split.after_time < since_moved[ax]))}  # type: ignore

        split_command_string = create_coords_list(split_target)
        primary_command_string = create_coords_list(moving_target)
        backlash_command_string = create_coords_list(backlash_target)

        self.dwell_axes(''.join(non_moving_axes))
        self.activate_axes(''.join(moving_axes))

        checked_speed = speed or self._combined_speed

        if split_command_string:
            # set fullstepping if necessary
            split_prefix, split_postfix = self._build_fullstep_configurations(
                ''.join((ax for ax in split_target.keys()
                         if self._move_split_config[ax].fullstep))
            )

            # move at the slowest required speed
            split_speed = min(split.split_speed
                              for ax, split in self._move_split_config.items()
                              if ax in split_target)

            # use the higher current from the split config without changing
            # our global cache
            split_prefix.add_builder(
                builder=self._build_speed_command(split_speed)
            )
            cached = {}
            for ax in split_target.keys():
                cached[ax] = self.current[ax]
                self.current[ax] = self._move_split_config[ax].split_current
            split_prefix.add_builder(
                builder=self._generate_current_command()
            )
            for ax in split_target.keys():
                self.current[ax] = cached[ax]

            split_command = _command_builder().add_gcode(
                gcode=GCODE.MOVE
            ).add_builder(
                builder=split_command_string
            )
        else:
            split_prefix = _command_builder()
            split_command = _command_builder()
            split_postfix = _command_builder()

        command = _command_builder()

        if split_command_string or (checked_speed != self._combined_speed):
            command.add_builder(
                builder=self._build_speed_command(checked_speed)
            )

        # introduce the standard currents
        command.add_builder(
            builder=self._generate_current_command()
        )

        if backlash_command_string:
            command.add_gcode(
                gcode=GCODE.MOVE
            ).add_builder(
                builder=backlash_command_string
            )

        command.add_gcode(
            GCODE.MOVE
        ).add_builder(
            builder=primary_command_string
        )
        if checked_speed != self._combined_speed:
            command.add_builder(
                builder=self._build_speed_command(self._combined_speed)
            )

        for axis in target.keys():
            self.engaged_axes[axis] = True
        if home_flagged_axes:
            self.home_flagged_axes(''.join(list(target.keys())))

        def _do_split():
            try:
                for sc in (c for c in (split_prefix, split_command) if c):
                    self._send_command(sc)
            finally:
                if split_postfix:
                    self._send_command(split_postfix)
        try:
            log.debug(f"move: {command}")
            # TODO (hmg) a movement's timeout should be calculated by
            # how long the movement is expected to take.
            _do_split()
            self._send_command(command, timeout=DEFAULT_EXECUTE_TIMEOUT)
        finally:
            # dwell pipette motors because they get hot
            plunger_axis_moved = ''.join(set('BC') & set(target.keys()))
            if plunger_axis_moved:
                self.dwell_axes(plunger_axis_moved)
                self._set_saved_current()
            self._axes_moved_at.mark_moved(moving_axes)

        self._update_position(target)

    def home(self,
             axis: str = AXES,
             disabled: str = DISABLE_AXES) -> Dict[str, float]:

        self.run_flag.wait()

        axis = axis.upper()

        # If Y is requested make sure we home X first
        if 'Y' in axis:
            axis += 'X'
        # If horizontal movement is requested, ensure we raise the instruments
        if 'X' in axis:
            axis += 'ZA'
        # These two additions are safe even if they duplicate requested axes
        # because of the use of set operations below, which will de-duplicate
        # characters from the resulting string

        # HOME_SEQUENCE defines a pattern for homing, specifically that the
        # ZABC axes should be homed first so that horizontal movement doesn't
        # happen with the pipette down (which could bump into things). Then
        # the X axis is homed, which has to happen before Y. Finally Y can be
        # homed. This variable will contain the sequence just explained, but
        # filters out unrequested axes using set intersection (&) and then
        # filters out disabled axes using set difference (-)
        home_sequence = list(filter(
            None,
            [
                ''.join(set(group) & set(axis) - set(disabled))
                for group in HOME_SEQUENCE
            ]))

        non_moving_axes = ''.join(
            ax for ax in AXES if ax not in home_sequence
        )
        self.dwell_axes(non_moving_axes)
        log.info(f"Homing axes {axis} in sequence {home_sequence}")
        for axes in home_sequence:
            if 'X' in axes:
                self._home_x()
            elif 'Y' in axes:
                self._home_y()
            else:
                # if we are homing neither the X nor Y axes, simple home
                self.activate_axes(axes)
                self._do_relative_splits_during_home_for(
                    ''.join(ax for ax in axes if ax in 'BC'))

                command = self._generate_current_command()
                command.add_gcode(
                    gcode=GCODE.HOME
                ).add_element(
                    element=''.join(sorted(axes))
                )
                try:
                    # home commands are executed before ack, use a long ack
                    # timeout and short execute timeout
                    self._send_command(
                        command, ack_timeout=DEFAULT_EXECUTE_TIMEOUT,
                        timeout=DEFAULT_ACK_TIMEOUT)
                    self.update_homed_flags(flags={ax: True for ax in axes})
                finally:
                    # always dwell an axis after it has been homed
                    self.dwell_axes(axes)
                    self._set_saved_current()

        # Only update axes that have been selected for homing
        homed_axes = "".join(home_sequence)
        homed = {
            ax: self.homed_position.get(ax)
            for ax in homed_axes
        }
        self.update_position(default=homed)

        for ax in homed_axes:
            self.engaged_axes[ax] = True

        # coordinate after homing might not sync with default in API
        # so update this driver's homed position using current coordinates
        new = {
            ax: self.position[ax]
            for ax in homed_axes
        }
        self._homed_position.update(new)
        self._axes_moved_at.mark_moved(homed_axes)
        return self.position

    def _build_fullstep_configurations(self, axes: str) \
            -> Tuple[CommandBuilder, CommandBuilder]:
        """ For one or more specified pipette axes,
        build a prefix and postfix command string that will configure
        the step mode and steps/mm value to
        - in the prefix: set full stepping with an appropriate steps/mm
        - in the postfix: set 1/32 microstepping with the correct steps/mm

        Prefix will always be empty or end with a space, and postfix will
        always be empty or start with a space, so they can be added to
        command strings easily
        """
        prefix = _command_builder()
        postfix = _command_builder()
        if not axes:
            return prefix, postfix
        assert all((ax in 'BC' for ax in axes)),\
            'only plunger axes have controllable microstepping'
        for ax in axes:
            prefix.add_gcode(gcode=MICROSTEPPING_GCODES[ax]['DISABLE'])
        for ax in axes:
            postfix.add_gcode(gcode=MICROSTEPPING_GCODES[ax]['ENABLE'])

        prefix.add_builder(builder=self._build_steps_per_mm({
            ax: self.steps_per_mm[ax] / 32
            for ax in axes
        })).add_gcode(
            gcode=GCODE.DWELL
        ).add_float(prefix='P', value=0.01, precision=None)

        postfix.add_builder(builder=self._build_steps_per_mm({
            ax: self.steps_per_mm[ax]
            for ax in axes
        })).add_gcode(
            gcode=GCODE.DWELL
        ).add_float(prefix='P', value=0.01, precision=None)
        return prefix, postfix

    def _do_relative_splits_during_home_for(self, axes: str):
        """ Handle split moves for unsticking axes before home.

        This is particularly ugly bit of code that flips the motor controller
        into relative mode since we don't necessarily ktime where we are.

        It will induce a movement. It should really only be called before a
        home because it doesn't update the position cache.

        :param axes: A string that is a sequence of plunger axis names.
        """
        assert all(ax.lower() in 'bc' for ax in axes),\
            'only plunger axes may be unstuck'
        since_moved = self._axes_moved_at.time_since_moved()
        split_currents = _command_builder().add_gcode(
            gcode=GCODE.SET_CURRENT
        )
        split_moves = _command_builder().add_gcode(
            gcode=GCODE.MOVE
        )
        applicable_speeds: List[float] = []
        log.debug(f"Finding splits for {axes} with since moved {since_moved}")
        to_unstick = [
            ax for ax in axes if
            (since_moved.get(ax) is None
             or (
             self._move_split_config.get(ax)
             and since_moved[ax]  # type: ignore
                 > self._move_split_config[ax].after_time))   # type: ignore
        ]
        for axis in axes:
            msc = self._move_split_config.get(axis)
            log.debug(f"axis {axis}: msc {msc}")
            if not msc:
                continue
            if axis in to_unstick:
                log.debug(f"adding unstick for {axis}")
                split_currents.add_float(
                    prefix=axis, value=msc.split_current, precision=None
                )
                split_moves.add_float(
                    prefix=axis, value=-msc.split_distance, precision=None
                )
                applicable_speeds.append(msc.split_speed)
        if not applicable_speeds:
            log.debug("no unstick needed")
            # nothing to do
            return

        fullstep_prefix, fullstep_postfix\
            = self._build_fullstep_configurations(
                ''.join(to_unstick))

        command_sequence = [
            fullstep_prefix.add_builder(
                builder=split_currents
            ).add_gcode(
                gcode=GCODE.DWELL
            ).add_float(
                prefix="P", value=CURRENT_CHANGE_DELAY, precision=None
            ).add_builder(
                builder=self._build_speed_command(min(applicable_speeds))
            ).add_gcode(gcode=GCODE.RELATIVE_COORDS),
            split_moves
            ]
        try:
            for command_string in command_sequence:
                self._send_command(
                    command_string, timeout=DEFAULT_EXECUTE_TIMEOUT,
                    suppress_home_after_error=True)
        except SmoothieError:
            pass
        finally:
            self._send_command(
                _command_builder().add_gcode(
                    gcode=GCODE.ABSOLUTE_COORDS
                ).add_builder(
                    builder=fullstep_postfix
                ).add_builder(
                    builder=self._build_speed_command(self._combined_speed))
            )

    def fast_home(self, axis, safety_margin):
        """ home after a controlled motor stall

        Given a ktimen distance we have just stalled along an axis, move
        that distance away from the homing switch. Then finish with home.
        """
        # move some mm distance away from the target axes endstop switch(es)
        destination = {
            ax: self.homed_position.get(ax) - abs(safety_margin)
            for ax in axis.upper()
        }

        # there is a chance the axis will hit it's home switch too soon
        # if this happens, catch the error and continue with homing afterwards
        try:
            self.move(destination)
        except SmoothieError:
            pass

        # then home once we're closer to the endstop(s)
        disabled = ''.join(ax for ax in AXES if ax not in axis.upper())
        return self.home(axis=axis, disabled=disabled)

    def unstick_axes(
            self, axes: str, distance: float = None, speed: float = None):
        """
        The plunger axes on OT2 can build up static friction over time and
        when it's cold. To get over this, the robot can move that plunger at
        normal current and a very slow speed to increase the torque, removing
        the static friction

        axes:
            String containing each axis to be moved. Ex: 'BC' or 'ZABC'

        distance:
            Distance to travel in millimeters (default is 1mm)

        speed:
            Millimeters-per-second to travel to travel at (default is 1mm/sec)
        """
        for ax in axes:
            if ax not in AXES:
                raise ValueError(f'Unknown axes: {axes}')

        if distance is None:
            distance = UNSTICK_DISTANCE
        if speed is None:
            speed = UNSTICK_SPEED

        self.push_active_current()
        self.set_active_current({
            ax: self._config.high_current['default'][ax]  # type: ignore
            for ax in axes
            })
        self.push_axis_max_speed()
        self.set_axis_max_speed({ax: speed for ax in axes})

        # only need to request switch state once
        state_of_switches = {ax: False for ax in AXES}
        if not self.simulating:
            state_of_switches = self.switch_state

        # incase axes is pressing endstop, home it slowly instead of moving
        homing_axes = ''.join(ax for ax in axes if state_of_switches[ax])
        moving_axes = {
            ax: self.position[ax] - distance  # retract
            for ax in axes
            if (not state_of_switches[ax]) and (ax not in homing_axes)
        }

        try:
            if moving_axes:
                self.move(moving_axes)
            if homing_axes:
                self.home(homing_axes)
        finally:
            self.pop_active_current()
            self.pop_axis_max_speed()

    def pause(self):
        if not self.simulating:
            self.run_flag.clear()

    def resume(self):
        if not self.simulating:
            self.run_flag.set()

    def delay(self, seconds: float):
        # per http://smoothieware.org/supported-g-codes:
        # In grbl mode P is float seconds to comply with gcode standards
        command = _command_builder().add_gcode(
            gcode=GCODE.DWELL
        ).add_float(
            prefix="P", value=seconds, precision=None
        )

        log.debug(f"delay: {command}")
        self._send_command(command, timeout=int(seconds) + 1)

    def probe_axis(
            self, axis: str, probing_distance: float) -> Dict[str, float]:
        if axis.upper() in AXES:
            self.engaged_axes[axis] = True
            command = _command_builder().add_gcode(
                gcode=GCODE.PROBE
            ).add_int(
                prefix="F", value=420   # 420 mm/min (7 mm/sec) to avoid resonance
            ).add_float(
                prefix=axis.upper(), value=probing_distance, precision=None
            )
            log.debug(f"probe_axis: {command}")
            try:
                self._send_command(
                    command=command, ack_timeout=DEFAULT_MOVEMENT_TIMEOUT,
                    suppress_home_after_error=True)
            except SmoothieError as se:
                log.exception("Tip probe failure")
                self.home(axis)
                if 'probe' in str(se).lower():
                    raise TipProbeError(se.ret_code, se.command)
                else:
                    raise
            self.update_position(self.position)
            return self.position
        else:
            raise RuntimeError(f"Cant probe axis {axis}")

    def turn_on_blue_button_light(self):
        self._gpio_chardev.set_button_light(blue=True)

    def turn_on_green_button_light(self):
        self._gpio_chardev.set_button_light(green=True)

    def turn_on_red_button_light(self):
        self._gpio_chardev.set_button_light(red=True)

    def turn_off_button_light(self):
        self._gpio_chardev.set_button_light(
            red=False, green=False, blue=False)

    def turn_on_rail_lights(self):
        self._gpio_chardev.set_rail_lights(True)

    def turn_off_rail_lights(self):
        self._gpio_chardev.set_rail_lights(False)

    def get_rail_lights_on(self):
        return self._gpio_chardev.get_rail_lights()

    def read_button(self):
        return self._gpio_chardev.read_button()

    def read_window_switches(self):
        return self._gpio_chardev.read_window_switches()

    def set_lights(self, button: bool = None, rails: bool = None):
        if button is not None:
            self._gpio_chardev.set_button_light(blue=button)
        if rails is not None:
            self._gpio_chardev.set_rail_lights(rails)

    def get_lights(self) -> Dict[str, bool]:
        return {'button': self._gpio_chardev.get_button_light()[2],
                'rails': self._gpio_chardev.get_rail_lights()}

    def kill(self):
        """
        In order to terminate Smoothie motion immediately (including
        interrupting a command in progress, we set the reset pin low and then
        back to high, then call `_setup` method to send the RESET_FROM_ERROR
        Smoothie code to return Smoothie to a normal waiting state and reset
        any other state needed for the driver.
        """
        log.debug("kill")
        self.hard_halt()
        self._reset_from_error()
        self._setup()

    def home_flagged_axes(self, axes_string: str):
        """
        Given a list of axes to check, this method will home each axis if
        Smoothieware's internal flag sets it as needing to be homed
        """
        axes_that_need_to_home = [
            axis
            for axis, already_homed in self.homed_flags.items()
            if (not already_homed) and (axis in axes_string)
        ]
        if axes_that_need_to_home:
            axes_string = ''.join(axes_that_need_to_home)
            self.home(axes_string)

    def _smoothie_reset(self):
        log.debug(f'Resetting Smoothie (simulating: {self.simulating})')
        if self.simulating:
            pass
        else:
            self._gpio_chardev.set_reset_pin(False)
            self._gpio_chardev.set_isp_pin(True)
            sleep(0.25)
            self._gpio_chardev.set_reset_pin(True)
            sleep(0.25)
            self._wait_for_ack()
            self._reset_from_error()

    def _smoothie_programming_mode(self):
        log.debug(f'Setting Smoothie to ISP mode (simulating: {self.simulating})')
        if self.simulating:
            pass
        else:
            self._gpio_chardev.set_reset_pin(False)
            self._gpio_chardev.set_isp_pin(False)
            sleep(0.25)
            self._gpio_chardev.set_reset_pin(True)
            sleep(0.25)
            self._gpio_chardev.set_isp_pin(True)
            sleep(0.25)

    def hard_halt(self):
        log.debug(f'Halting Smoothie (simulating: {self.simulating}')
        if self.simulating:
            pass
        else:
            self._is_hard_halting.set()
            self._gpio_chardev.set_halt_pin(False)
            sleep(0.25)
            self._gpio_chardev.set_halt_pin(True)
            sleep(0.25)
            self.run_flag.set()

    async def update_firmware(self,  # noqa: C901
                              filename: str,
                              loop: asyncio.AbstractEventLoop = None,
                              explicit_modeset: bool = True) -> str:
        """
        Program the smoothie board with a given hex file.

        If explicit_modeset is True (default), explicitly place the smoothie in
        programming mode.

        If explicit_modeset is False, assume the smoothie is already in
        programming mode.
        """
        try:
            smoothie_update._ensure_programmer_executable()
        except OSError as ose:
            if ose.errno == 30:
                # This is "read only filesystem" and happens on buildroot
                pass
            else:
                raise

        if not self.is_connected():
            log.info("Getting port to connect")
            self._connect_to_port()

        assert self._connection,\
            'driver must have been initialized with a port'
        # get port name
        port = self._connection.port

        if explicit_modeset:
            log.info("Setting programming mode")
            # set smoothieware into programming mode
            self._smoothie_programming_mode()
            # close the port so other application can access it
            self._connection.close()

        # run lpc21isp, THIS WILL TAKE AROUND 1 MINUTE TO COMPLETE
        update_cmd = f'lpc21isp -wipe -donotstart {filename} ' \
                     f'{port} {self._config.serial_speed} 12000'
        kwargs: Dict[str, Any] = {
            'stdout': asyncio.subprocess.PIPE,
            'stderr': asyncio.subprocess.PIPE}
        if loop:
            kwargs['loop'] = loop
        log.info(update_cmd)
        before = time()
        proc = await asyncio.create_subprocess_shell(
            update_cmd, **kwargs)
        created = time()
        log.info(f"created lpc21isp subproc in {created-before}")
        out_b, err_b = await proc.communicate()
        done = time()
        log.info(f"ran lpc21isp subproc in {done-created}")
        if proc.returncode != 0:
            log.error(
                f"Smoothie update failed: {proc.returncode}"
                f" {out_b!r} {err_b!r}")
            raise RuntimeError(
                f"Failed to program smoothie: {proc.returncode}: {err_b!r}")
        else:
            log.info("Smoothie update complete")
        try:
            self._connection.close()
        except Exception:
            log.exception('Failed to close smoothie connection.')
        # re-open the port
        self._connection.open()
        # reset smoothieware
        self._smoothie_reset()
        # run setup gcodes
        self._setup()

        return out_b.decode().strip()

    # ----------- END Public interface ------------ #



"""
USB Simulating Driver.

A class to convert info from the usb bus into a
more readable format.
"""
from typing import List, Set, Union

from opentrons.hardware_control.modules.types import ModuleAtPort
from opentrons.hardware_control.types import BoardRevision

from .interfaces import USBDriverInterface
from .types import USBPort


class USBBusSimulator(USBDriverInterface):
    def __init__(self, board_revision: BoardRevision):
        self._usb_dev: List[USBPort] = self.read_usb_bus()
        self._sorted: Set[Union[int, str]] = set()
        self._board_revision = board_revision

    @staticmethod
    def read_bus() -> List[str]:
        """
        Read the USB Bus information.

        Use the sys bus path to find all of the USBs with
        active devices connected to them.
        """
        return ['']

    @staticmethod
    def read_symlink(virtual_port: str) -> str:
        """
        """
        return ''

    @property
    def board_revision(self) -> BoardRevision:
        return self._board_revision

    @property
    def usb_dev(self) -> List[USBPort]:
        """
        USBBus property: usb_dev.

        :returns: The list of ports found from
        the usb bus.
        """
        return self._usb_dev

    @usb_dev.setter
    def usb_dev(self, ports: List[USBPort]) -> None:
        """
        USBBus setter: usb_dev.

        :param ports: The list of ports found from
        the usb bus.
        """
        self._usb_dev = ports

    @property
    def sorted_ports(self) -> Set:
        """
        USBBus property: sorted_ports.

        :returns: The set of sorted ports
        """
        return self._sorted

    @sorted_ports.setter
    def sorted_ports(self, sorted: Set) -> None:
        """
        USBBus setter: sorted_ports.

        :param sorted: The updated set of usb ports.
        """
        self._sorted = sorted

    def read_usb_bus(self) -> List[USBPort]:
        """
        Read usb bus

        Take the value returned from the USB bus and match
        the paths to the expected port paths for modules.
        :returns: A list of matching ports as dataclasses
        """
        return []

    def find_port(self, device_path: str) -> USBPort:
        """
        Find port.

        Take the value returned from the USB bus and match
        the paths to the expected port paths for modules.
        :param device_path: The device path of a module, which
        generally contains tty/tty* in its name.
        :returns: The matching port, or an empty port dataclass
        """
        return USBPort(
            name='', sub_names=[], device_path=device_path)

    def sort_ports(self) -> None:
        pass

    def match_virtual_ports(
            self, virtual_port: List[ModuleAtPort]
            ) -> List[ModuleAtPort]:
        return virtual_port



import logging
import numpy as np  # type: ignore
from dataclasses import dataclass
from typing import Optional, List

from opentrons import config

from opentrons.config.robot_configs import get_legacy_gantry_calibration
from opentrons.calibration_storage import modify, types, get
from opentrons.types import Mount
from opentrons.util import linal

from .util import DeckTransformState

log = logging.getLogger(__name__)


@dataclass
class RobotCalibration:
    deck_calibration: types.DeckCalibration


def build_temporary_identity_calibration() -> RobotCalibration:
    """
    Get a temporary identity deck cal suitable for use during
    calibration
    """
    return RobotCalibration(
        deck_calibration=types.DeckCalibration(
            attitude=linal.identity_deck_transform().tolist(),
            source=types.SourceType.default,
            status=types.CalibrationStatus()))


def validate_attitude_deck_calibration(deck_cal: types.DeckCalibration):
    """
    This function determines whether the deck calibration is valid
    or not based on the following use-cases:

    TODO(lc, 8/10/2020): Expand on this method, or create
    another method to diagnose bad instrument offset data
    """
    curr_cal = np.array(deck_cal.attitude)
    row, _ = curr_cal.shape
    rank = np.linalg.matrix_rank(curr_cal)
    if row != rank:
        # Check that the matrix is non-singular
        return DeckTransformState.SINGULARITY
    elif not deck_cal.last_modified:
        # Check that the matrix is not an identity
        return DeckTransformState.IDENTITY
    else:
        # Transform as it stands is sufficient.
        return DeckTransformState.OK


def validate_gantry_calibration(gantry_cal: List[List[float]]):
    """
    This function determines whether the gantry calibration is valid
    or not based on the following use-cases:
    """
    curr_cal = np.array(gantry_cal)
    row, _ = curr_cal.shape

    rank = np.linalg.matrix_rank(curr_cal)

    id_matrix = linal.identity_deck_transform()

    z = abs(curr_cal[2][-1])

    outofrange = z < 16 or z > 34
    if row != rank:
        # Check that the matrix is non-singular
        return DeckTransformState.SINGULARITY
    elif np.array_equal(curr_cal, id_matrix):
        # Check that the matrix is not an identity
        return DeckTransformState.IDENTITY
    elif outofrange:
        # Check that the matrix is not out of range.
        return DeckTransformState.BAD_CALIBRATION
    else:
        # Transform as it stands is sufficient.
        return DeckTransformState.OK


def migrate_affine_xy_to_attitude(
        gantry_cal: List[List[float]]) -> types.AttitudeMatrix:
    masked_transform = np.array([
        [True, True, True, False],
        [True, True, True, False],
        [False, False, False, False],
        [False, False, False, False]])
    masked_array = np.ma.masked_array(gantry_cal, ~masked_transform)
    attitude_array = np.zeros((3, 3))
    np.put(attitude_array, [0, 1, 2], masked_array[0].compressed())
    np.put(attitude_array, [3, 4, 5], masked_array[1].compressed())
    np.put(attitude_array, 8, 1)
    return attitude_array.tolist()


def save_attitude_matrix(
        expected: linal.SolvePoints, actual: linal.SolvePoints,
        pipette_id: str, tiprack_hash: str):
    attitude = linal.solve_attitude(expected, actual)
    modify.save_robot_deck_attitude(attitude, pipette_id, tiprack_hash)


def load_attitude_matrix() -> types.DeckCalibration:
    calibration_data = get.get_robot_deck_attitude()
    gantry_cal = get_legacy_gantry_calibration()
    if not calibration_data and gantry_cal:
        if validate_gantry_calibration(gantry_cal) == DeckTransformState.OK:
            log.debug(
                "Attitude deck calibration matrix not found. Migrating "
                "existing affine deck calibration matrix to {}".format(
                    config.get_opentrons_path('robot_calibration_dir')))
            attitude = migrate_affine_xy_to_attitude(gantry_cal)
            modify.save_robot_deck_attitude(transform=attitude,
                                            pip_id=None,
                                            lw_hash=None,
                                            source=types.SourceType.legacy)
            calibration_data = get.get_robot_deck_attitude()

    if calibration_data:
        return calibration_data
    else:
        # load default if deck calibration data do not exist
        return types.DeckCalibration(
            attitude=config.robot_configs.DEFAULT_DECK_CALIBRATION_V2,
            source=types.SourceType.default,
            status=types.CalibrationStatus())


def load_pipette_offset(
        pip_id: Optional[str],
        mount: Mount) -> types.PipetteOffsetByPipetteMount:
    # load default if pipette offset data do not exist
    pip_cal_obj = types.PipetteOffsetByPipetteMount(
        offset=config.robot_configs.DEFAULT_PIPETTE_OFFSET,
        source=types.SourceType.default,
        status=types.CalibrationStatus())
    if pip_id:
        pip_offset_data = get.get_pipette_offset(pip_id, mount)
        if pip_offset_data:
            return pip_offset_data
    return pip_cal_obj


def load() -> RobotCalibration:
    return RobotCalibration(
        deck_calibration=load_attitude_matrix())



import re
from dataclasses import dataclass
from typing import Dict, Generator, Optional


@dataclass
class Command:
    gcode: str
    body: str
    params: Dict[str, Optional[float]]


class Parser:
    """Gcode line parser."""

    # TODO (al, 2021-05-11): Should G01 and G1 be treated differently? Currently
    #  they are two different Gcodes.
    GCODE_RE = re.compile(r"(?:(?:[MG]\d+\.?\d*)|dfu|version)")
    """A gcode is either M or G followed by decimal. Or 'dfu' or 'version'."""
    ALPHA_PREFIXED_NUMBER_RE = re.compile(r"(?P<prefix>[A-Z])(?P<number>-?\d*\.?\d*)")
    """All parameters are a capital letter followed by a decimal value."""

    def parse(self, line: str) -> Generator[Command, None, None]:
        """
        Parse a line to extract commands.

        Args:
            line: a line containing gcodes and their values. spaces are optional.

        Returns:
            Command object
        """
        line = line.strip()
        previous = None
        for i in self.GCODE_RE.finditer(line):
            if previous:
                yield self._create_command(
                    line[previous.start(): previous.end()],
                    line[previous.end(): i.start()]
                )
            else:
                # This is the first match. It better be at the beginning or
                # there's junk in the beginning.
                if i.start() != 0:
                    raise ValueError(f"Invalid content: {line}")
            previous = i
        if previous:
            # Create command from final GCODE and remainder of the line.
            yield self._create_command(
                line[previous.start(): previous.end()],
                line[previous.end():]
            )
        elif line:
            # There are no GCODEs and it's a non-empty line.
            raise ValueError(f"Invalid content: {line}")

    @staticmethod
    def _create_command(gcode: str, body: str) -> Command:
        """
        Create a Command.

        Args:
            gcode: the gcode
            body: the parameter string

        Returns: a Command object
        """
        pars = (i.groupdict() for i in Parser.ALPHA_PREFIXED_NUMBER_RE.finditer(body))
        return Command(
            gcode=gcode,
            body=body.strip(),
            params={
                p['prefix']: float(p['number']) if p['number'] else None for p in pars
            }
        )



import asyncio
import logging
import os
import threading
import serial  # type: ignore
from queue import Queue
try:
    import select
except ModuleNotFoundError:
    select = None  # type: ignore
from time import sleep
from collections import deque
from typing import Callable, Optional, Mapping, Tuple, Deque, TYPE_CHECKING
from serial.serialutil import SerialException  # type: ignore
from opentrons.drivers import serial_communication, utils
from opentrons.drivers.serial_communication import SerialNoResponse

if TYPE_CHECKING:
    # avoid an issue where Queue doesn't support generics at runtime
    CommandQueue = Queue[Tuple[str, Callable[[str], None]]]
else:
    CommandQueue = Queue


log = logging.getLogger(__name__)

GCODES = {
    'OPEN_LID': 'M126',
    'CLOSE_LID': 'M127',
    'GET_LID_STATUS': 'M119',
    'SET_LID_TEMP': 'M140',
    'GET_LID_TEMP': 'M141',
    'EDIT_PID_PARAMS': 'M301',
    'SET_PLATE_TEMP': 'M104',
    'GET_PLATE_TEMP': 'M105',
    'SET_RAMP_RATE': 'M566',
    'DEACTIVATE_ALL': 'M18',
    'DEACTIVATE_LID': 'M108',
    'DEACTIVATE_BLOCK': 'M14',
    'DEVICE_INFO': 'M115'
}
LID_TARGET_DEFAULT = 105.0    # Degree celsius (floats)
LID_TARGET_MIN = 37.0
LID_TARGET_MAX = 110.0
BLOCK_TARGET_MIN = 0.0
BLOCK_TARGET_MAX = 99.0
TEMP_UPDATE_RETRIES = 50
TEMP_BUFFER_MAX_LEN = 10


def _build_temp_code(temp: float,
                     hold_time: Optional[float],
                     volume: Optional[float]):
    if temp < BLOCK_TARGET_MIN:
        temp = BLOCK_TARGET_MIN
    if temp > BLOCK_TARGET_MAX:
        temp = BLOCK_TARGET_MAX
    cmd = f"{GCODES['SET_PLATE_TEMP']} S{temp}"
    if hold_time:
        cmd += f' H{hold_time}'
    if volume:
        cmd += f' V{volume}'
    return cmd, temp


TC_BAUDRATE = 115200
TC_BOOTLOADER_BAUDRATE = 1200
# TODO (Laura 20190327) increased the thermocycler command timeout
# temporarily until we can change the firmware to asynchronously handle
# the lid being open and closed
SERIAL_ACK = '\r\n'
TC_COMMAND_TERMINATOR = SERIAL_ACK
TC_ACK = 'ok' + SERIAL_ACK + 'ok' + SERIAL_ACK
ERROR_KEYWORD = 'error'
DEFAULT_TC_TIMEOUT = 40
DEFAULT_COMMAND_RETRIES = 3
DEFAULT_STABILIZE_DELAY = 0.1
DEFAULT_POLLER_WAIT_SECONDS = 0.1
POLLING_FREQUENCY_MS = 1000
HOLD_TIME_FUZZY_SECONDS = POLLING_FREQUENCY_MS / 1000 * 5
TEMP_THRESHOLD = 0.3


class ThermocyclerError(Exception):
    pass


class SimulatingDriver:
    def __init__(self, sim_model: str = None):
        self._target_temp: Optional[float] = None
        self._ramp_rate: Optional[float] = None
        self._hold_time: Optional[float] = None
        self._active = False
        self._port = None
        self._lid_status = 'open'
        self._lid_target: Optional[float] = None
        self._lid_heating_active = False

    async def open(self):
        self._lid_status = 'open'
        return self._lid_status

    async def close(self):
        self._lid_status = 'closed'
        return self._lid_status

    @property
    def status(self):
        return 'holding at target' if self._active else 'idle'

    @property
    def lid_status(self):
        return self._lid_status

    @property
    def ramp_rate(self):
        return self._ramp_rate

    @property
    def hold_time(self):
        # Simulating driver acts as if cycles end immediately
        return 0

    @property
    def temperature(self):
        return self._target_temp

    @property
    def target(self):
        return self._target_temp

    @property
    def lid_target(self):
        return self._lid_target

    @property
    def lid_temp_status(self):
        return 'holding at target' if self._lid_heating_active else 'idle'

    @property
    def lid_temp(self):
        return self._lid_target

    async def connect(self, port):
        self._port = port

    def disconnect(self):
        self._port = None

    async def set_temperature(self,
                              temp: float,
                              hold_time: float = None,
                              ramp_rate: float = None,
                              volume: float = None) -> None:
        self._target_temp = temp
        self._hold_time = hold_time
        self._ramp_rate = ramp_rate
        self._active = True

    async def set_lid_temperature(self, temp: Optional[float]):
        """ Set the lid temperature in deg Celsius """
        self._lid_heating_active = True
        self._lid_target = temp

    async def deactivate_lid(self):
        self._lid_heating_active = False
        self._lid_target = None

    async def deactivate_block(self):
        self._target_temp = None
        self._ramp_rate = None
        self._hold_time = None
        self._active = False

    async def deactivate_all(self):
        self._target_temp = None
        self._ramp_rate = None
        self._hold_time = None
        self._active = False
        self._lid_heating_active = False
        self._lid_target = None

    async def get_device_info(self):
        return {'serial': 'dummySerialTC',
                'model': 'dummyModelTC',
                'version': 'dummyVersionTC'}

    async def enter_programming_mode(self):
        pass


class TCPoller(threading.Thread):
    POLLING_FD_PATH = '/var/run/'

    def __init__(self, port, interrupt_callback, temp_status_callback,
                 lid_status_callback, lid_temp_status_callback):
        if not select:
            raise RuntimeError("Cannot connect to a Thermocycler from Windows")
        self._port = port
        self._connection = self._connect_to_port()
        self._interrupt_callback = interrupt_callback
        self._temp_status_callback = temp_status_callback
        self._lid_status_callback = lid_status_callback
        self._lid_temp_status_callback = lid_temp_status_callback
        self._lock = threading.Lock()
        self._command_queue: CommandQueue = Queue()

        # Note: the options and order of operations for opening file
        # descriptors is very specific. For more info, see:
        # http://pubs.opengroup.org/onlinepubs/007908799/xsh/open.html
        self._send_path = os.path.join(
            self.POLLING_FD_PATH, f"tc_send_fifo_{hash(self)}")
        os.mkfifo(self._send_path)
        send_read_fd = os.open(
            self._send_path, flags=os.O_RDONLY | os.O_NONBLOCK)
        self._send_read_file = os.fdopen(send_read_fd, 'rb')
        self._send_write_fd = open(self._send_path, 'wb', buffering=0)

        self._halt_path = os.path.join(
            self.POLLING_FD_PATH, f"tc_halt_fifo_{hash(self)}")
        os.mkfifo(self._halt_path)
        halt_read_fd = os.open(
            self._halt_path, flags=os.O_RDONLY | os.O_NONBLOCK)
        self._halt_read_file = os.fdopen(halt_read_fd, 'rb')
        self._halt_write_fd = open(self._halt_path, 'wb', buffering=0)

        self._poller = select.poll()
        self._poller.register(self._send_read_file, select.POLLIN)
        self._poller.register(self._halt_read_file, select.POLLIN)
        self._poller.register(self._connection, select.POLLIN)
        serial_thread_name = 'tc_serial_poller_{}'.format(hash(self))
        super().__init__(target=self._serial_poller, name=serial_thread_name)
        log.info("Starting TC thread {}".format(serial_thread_name))
        super().start()

    @property
    def port(self):
        return self._port

    def _serial_poller(self):
        """ Priority-sorted list of checks

        Highest priority is the 'halt' channel, which is used to kill the
        thread and the serial communication channel and allow everything to be
        cleaned up.

        Second is the lid-open interrupt, which should trigger a callback
        (typically to halt the robot).

        Third is an enqueued command to send to the Thermocycler.

        Fourth (if no other work is available) is to query the Thermocycler for
        its current temp, target temp, and time remaining in its current cycle.
        """
        while True:
            _next = dict(self._poller.poll(POLLING_FREQUENCY_MS))
            if self._halt_read_file.fileno() in _next:
                log.debug("Poller [{}]: halt".format(hash(self)))
                self._halt_read_file.read()
                # Note: this is discarded because we send a set message to halt
                # the thread--don't currently need to parse it
                break

            elif self._connection.fileno() in _next:
                # Lid-open interrupt
                log.debug("Poller [{}]: interrupt".format(hash(self)))
                res = self._connection.read_until(SERIAL_ACK)
                self._interrupt_callback(res)

            elif self._send_read_file.fileno() in _next:
                self._send_read_file.read(1)
                command, callback = self._command_queue.get()
                log.debug("Poller [{}]: send {}".format(hash(self), command))
                res = self._send_command(command)
                callback(res)
            else:
                # Nothing else to do--update device status
                log.debug("Poller [{}]: updating temp".format(hash(self)))
                res = self._send_command(GCODES['GET_PLATE_TEMP'])
                self._temp_status_callback(res)
                res = self._send_command(GCODES['GET_LID_STATUS'])
                self._lid_status_callback(res)
                res = self._send_command(GCODES['GET_LID_TEMP'])
                self._lid_temp_status_callback(res)
        log.info("Exiting TC poller loop [{}]".format(hash(self)))

    def _wait_for_ack(self):
        """
        This method writes a sequence of newline characters, which will
        guarantee the device responds with 'ok\r\nok\r\n' within 1 second
        """
        self._send_command(SERIAL_ACK, timeout=DEFAULT_TC_TIMEOUT)

    def _send_command(self, command, timeout=DEFAULT_TC_TIMEOUT):
        command_line = command + ' ' + TC_COMMAND_TERMINATOR
        ret_code = self._recursive_write_and_return(
            command_line, timeout, DEFAULT_COMMAND_RETRIES)
        if ERROR_KEYWORD in ret_code.lower():
            log.error('Received error message from Thermocycler: {}'.format(
                ret_code))
            raise ThermocyclerError(ret_code)
        return ret_code.strip()

    def _recursive_write_and_return(self, cmd, timeout, retries):
        try:
            return serial_communication.write_and_return(
                cmd, TC_ACK, self._connection, timeout,
                tag=f'thermocycler {id(self)}')
        except SerialNoResponse as e:
            retries -= 1
            if retries <= 0:
                raise e
            sleep(DEFAULT_STABILIZE_DELAY)
            if self._connection:
                self._connection.close()
                self._connection.open()
            return self._recursive_write_and_return(
                cmd, timeout, retries)

    def _connect_to_port(self):
        try:
            return serial_communication.connect(port=self._port,
                                                baudrate=TC_BAUDRATE)
        except SerialException:
            raise SerialException(
                "Thermocycler device not found on {}".format(self._port))

    def send(self, command, callback):
        with self._lock:
            self._command_queue.put((command, callback))
            self._send_write_fd.write(b'c')

    def close(self):
        log.debug("Halting TCPoller")
        self._halt_write_fd.write(b'q')

    def __del__(self):
        """ Clean up thread fifos"""
        log.debug("Cleaning up thread fifos in TCPoller.")
        try:
            os.unlink(self._send_path)
        except NameError:
            pass
        try:
            os.unlink(self._halt_path)
        except NameError:
            pass


class Thermocycler:
    def __init__(self, interrupt_callback):
        self._poller = None
        self._update_thread = None
        self._current_temp = None
        self._target_temp = None
        self._ramp_rate = None
        self._hold_time = None
        self._lid_status = None
        self._interrupt_cb = interrupt_callback
        self._lid_target = None
        self._lid_temp = None
        # to store previous _current_temp values:
        self._block_temp_buffer: Deque = deque(maxlen=TEMP_BUFFER_MAX_LEN)

    async def connect(self, port: str) -> 'Thermocycler':
        self.disconnect()
        self._poller = TCPoller(
            port, self._interrupt_callback,
            self._temp_status_update_callback,
            self._lid_status_update_callback,
            self._lid_temp_status_callback)

        # Check initial device lid state
        _lid_status_res = await self._write_and_wait(GCODES['GET_LID_STATUS'])
        if _lid_status_res:
            self._lid_status = utils.parse_string_value_from_substring(
                _lid_status_res.split()[-1])
        return self

    def disconnect(self) -> 'Thermocycler':
        if self.is_connected() or self._poller:
            self._poller.close()  # type: ignore
            self._poller.join()  # type: ignore
            log.debug("TC poller stopped.")
        self._poller = None
        return self

    async def deactivate_all(self):
        await self._write_and_wait(GCODES['DEACTIVATE_ALL'])

    async def deactivate_lid(self) -> None:
        await self._write_and_wait(GCODES['DEACTIVATE_LID'])

    async def deactivate_block(self):
        await self._write_and_wait(GCODES['DEACTIVATE_BLOCK'])

    def is_connected(self) -> bool:
        if not self._poller:
            return False
        return self._poller.is_alive()

    async def open(self):
        await self._write_and_wait(GCODES['OPEN_LID'])
        self.lid_status = 'open'
        return self.lid_status

    async def close(self):
        await self._write_and_wait(GCODES['CLOSE_LID'])
        self.lid_status = 'closed'
        return self.lid_status

    def hold_time_probably_set(self, new_hold_time: Optional[float]) -> bool:
        """
        Since we can only get hold time *remaining* from TC, by the time we
        read hold_time after a set_temperature, the hold_time in TC could have
        started counting down. So instead of checking for equality, we will
        have to check if the hold_time returned from TC is within a few seconds
        of the new hold time. The number of seconds is determined by status
        polling frequency.
        """
        if new_hold_time is None:
            return True
        if self._hold_time is None:
            return False
        lower_bound = max(0.0, new_hold_time - HOLD_TIME_FUZZY_SECONDS)
        return lower_bound <= self._hold_time <= new_hold_time

    async def set_temperature(self,
                              temp: float,
                              hold_time: float = None,
                              ramp_rate: float = None,
                              volume: float = None) -> None:
        if ramp_rate:
            ramp_cmd = f"{GCODES['SET_RAMP_RATE']} S{ramp_rate}"
            await self._write_and_wait(ramp_cmd)

        temp_cmd, temp = _build_temp_code(temp=temp,
                                          hold_time=hold_time,
                                          volume=volume)
        await self._write_and_wait(temp_cmd)
        retries = 0
        while self._target_temp != temp or \
                not self.hold_time_probably_set(hold_time):
            # Wait for the poller to update
            await asyncio.sleep(DEFAULT_POLLER_WAIT_SECONDS)
            retries += 1
            if retries > TEMP_UPDATE_RETRIES:
                raise ThermocyclerError(f'Thermocycler driver set the block '
                                        f'temp to T={temp} & H={hold_time} '
                                        f'but status reads '
                                        f'T={self._target_temp} & '
                                        f'H={self._hold_time}')

    async def set_lid_temperature(self, temp: float) -> None:
        if temp is None:
            _lid_target = LID_TARGET_DEFAULT
        else:
            if temp < LID_TARGET_MIN:
                _lid_target = LID_TARGET_MIN
            elif temp > LID_TARGET_MAX:
                _lid_target = LID_TARGET_MAX
            else:
                _lid_target = temp

        lid_temp_cmd = '{} S{}'.format(GCODES['SET_LID_TEMP'], _lid_target)
        await self._write_and_wait(lid_temp_cmd)
        retries = 0
        while self._lid_target != _lid_target:
            # Wait for the poller to update
            await asyncio.sleep(DEFAULT_POLLER_WAIT_SECONDS)
            retries += 1
            if retries > TEMP_UPDATE_RETRIES:
                raise ThermocyclerError(f'Thermocycler driver set lid temp to'
                                        f' {_lid_target} but self._lid_target'
                                        f' reads {self._lid_target}')

    def _lid_status_update_callback(self, lid_response):
        if lid_response:
            self._lid_status = utils.parse_string_value_from_substring(
                lid_response.split()[-1])

    def _temp_status_update_callback(self, temperature_response):
        # Payload is shaped like `T:95.0 C:77.4 H:600` where T is the
        # target temperature, C is the current temperature, and H is the
        # hold time remaining
        temp = utils.parse_plate_temperature_response(
            temperature_string=temperature_response,
            rounding_val=utils.TC_GCODE_ROUNDING_PRECISION
        )
        self._current_temp = temp.current
        self._target_temp = temp.target
        self._hold_time = temp.hold
        self._block_temp_buffer.append(self._current_temp)

    def _lid_temp_status_callback(self, lid_temp_res):
        # Payload is shaped like `T:95.0 C:77.4` where T is the
        # target temperature, C is the current temperature
        temp = utils.parse_temperature_response(
            temperature_string=lid_temp_res,
            rounding_val=utils.TC_GCODE_ROUNDING_PRECISION
        )
        self._lid_temp = temp.current
        self._lid_target = temp.target

    def _interrupt_callback(self, interrupt_response):
        # TODO sanitize response and then call the callback
        parsed_response = interrupt_response
        self._interrupt_cb(parsed_response)

    @property
    def temperature(self):
        return self._current_temp

    @property
    def target(self):
        return self._target_temp

    @property
    def hold_time(self):
        return self._hold_time

    @property
    def ramp_rate(self):
        return self._ramp_rate

    @property
    def lid_temp_status(self):
        if self.lid_temp is None:
            _status = 'error'
        if self.lid_target is None:
            _status = 'idle'
        else:
            diff = self.lid_target - self.lid_temp
            if abs(diff) < TEMP_THRESHOLD:
                _status = 'holding at target'
            elif diff < 0:
                _status = 'idle'  # TC lid can't actively cool
            else:
                _status = 'heating'
        return _status

    @property
    def status(self):
        if self.temperature is None:
            _status = 'error'
        elif self.target is None:
            _status = 'idle'
        else:
            diff = self.target - self.temperature
            if self._is_holding_at_target():
                _status = 'holding at target'
            elif diff < 0:
                _status = 'cooling'
            else:
                _status = 'heating'
        return _status

    def _is_holding_at_target(self) -> bool:
        """
        Checks block temp history to determine if block temp has stabilized at
        the target temperature. Returns true only if all values in history are
        within threshold range of target temperature.
        """
        if len(self._block_temp_buffer) < TEMP_BUFFER_MAX_LEN:
            # Not enough temp history
            return False
        else:
            return all(abs(self.target - t) < TEMP_THRESHOLD
                       for t in self._block_temp_buffer)

    @property
    def port(self) -> Optional[str]:
        if not self._poller:
            return None
        return self._poller.port

    @property
    def lid_status(self):
        return self._lid_status

    @lid_status.setter
    def lid_status(self, status):
        self._lid_status = status

    @property
    def lid_temp(self):
        return self._lid_temp

    @property
    def lid_target(self):
        return self._lid_target

    async def get_device_info(self) -> Mapping[str, str]:
        _device_info_res = await self._write_and_wait(GCODES['DEVICE_INFO'])
        if _device_info_res:
            return utils.parse_device_information(_device_info_res)
        else:
            raise ThermocyclerError("Thermocycler did not return device info")

    async def _write_and_wait(self, command):
        ret = None

        def cb(cmd):
            nonlocal ret
            ret = cmd

        assert self._poller, 'not connected'
        self._poller.send(command, cb)

        while None is ret:
            await asyncio.sleep(0.05)
            pass
        return ret

    async def enter_programming_mode(self):
        trigger_connection = serial.Serial(
            self.port, TC_BOOTLOADER_BAUDRATE, timeout=1)
        await asyncio.sleep(0.05)
        trigger_connection.close()
        self.disconnect()



import binascii
import logging
import time
from typing import Dict, Optional, Mapping, Iterable, Sequence
import re

from opentrons.drivers.types import Temperature, PlateTemperature

log = logging.getLogger(__name__)

# Number of digits after the decimal point for temperatures being sent
# to/from Temp-Deck
TEMPDECK_GCODE_ROUNDING_PRECISION = 0
TC_GCODE_ROUNDING_PRECISION = 2


KEY_VALUE_REGEX = re.compile(r"((?P<key>\S+):(?P<value>\S+))")


class ParseError(Exception):
    def __init__(self, error_message: str, parse_source: str) -> None:
        self.error_message = error_message
        self.parse_source = parse_source
        super().__init__(
            f"ParseError(error_message={error_message}, parse_source={parse_source})"
        )


def parse_string_value_from_substring(substring: str) -> str:
    """
    Returns the ascii value in the expected string "N:aa11bb22", where "N" is
    the key, and "aa11bb22" is string value to be returned
    """
    try:
        value = substring.split(':')[1]
        return str(value)
    except (ValueError, IndexError, TypeError, AttributeError):
        log.exception('Unexpected arg to parse_string_value_from_substring:')
        raise ParseError(
            error_message='Unexpected arg to parse_string_value_from_substring',
            parse_source=substring
        )


def parse_temperature_response(
        temperature_string: str, rounding_val: int
) -> Temperature:
    """Example input: "T:none C:25"""
    data = parse_key_values(temperature_string)
    try:
        return Temperature(
            current=parse_optional_number(data['C'], rounding_val),
            target=parse_optional_number(data['T'], rounding_val)
        )
    except KeyError:
        raise ParseError(
            error_message='Unexpected argument to parse_temperature_response',
            parse_source=temperature_string
        )


def parse_plate_temperature_response(
        temperature_string: str, rounding_val: int
) -> PlateTemperature:
    """Example input: "T:none C:25 H:123"""
    data = parse_key_values(temperature_string)
    try:
        return PlateTemperature(
            current=parse_optional_number(data['C'], rounding_val),
            target=parse_optional_number(data['T'], rounding_val),
            hold=parse_optional_number(data['H'], rounding_val)
        )
    except KeyError:
        raise ParseError(
            error_message='Unexpected argument to parse_lid_temperature_response',
            parse_source=temperature_string
        )


def parse_device_information(
        device_info_string: str) -> Dict[str, str]:
    """
    Parse the modules's device information response.

    Example response from temp-deck: "serial:aa11 model:bb22 version:cc33"
    """
    res = parse_key_values(device_info_string)

    try:
        return {key: res[key] for key in ['model', 'version', 'serial']}
    except KeyError as e:
        raise ParseError(
            error_message=f"Missing key '{str(e)}' in parse_device_information",
            parse_source=device_info_string
        )


def parse_key_values(value: str) -> Dict[str, str]:
    """Convert string in the format:
        'key1:value1 key2:value2'
        to dict
        {'key1': 'value1', 'key2': 'value2'}
    """
    res = {
        g.groupdict()['key']: g.groupdict()['value']
        for g in KEY_VALUE_REGEX.finditer(value)
    }
    return res


def parse_optional_number(value: str, rounding_val) -> Optional[float]:
    """Convert number to float. 'none' will be converted to None"""
    return None if value == "none" else parse_number(value, rounding_val)


def parse_number(value: str, rounding_val: int) -> float:
    """Convert string to float."""
    try:
        return round(float(value), rounding_val)
    except ValueError:
        raise ParseError(
            error_message='Unexpected argument to parse_number',
            parse_source=value
        )


class AxisMoveTimestamp:
    """ Keeps track of the last time axes were known to move """

    def __init__(self, axis_iter: Sequence[str]):
        self._moved_at: Dict[str, Optional[float]] = {
            ax: None for ax in axis_iter
        }

    def mark_moved(self, axis_iter: Sequence[str]):
        """ Indicate that a set of axes just moved """
        now = time.monotonic()
        self._moved_at.update({ax: now for ax in axis_iter})

    def time_since_moved(self) -> Mapping[str, Optional[float]]:
        """ Get a mapping of the time since each known axis moved """
        now = time.monotonic()
        return {ax: now-val if val else None
                for ax, val, in self._moved_at.items()}

    def reset_moved(self, axis_iter: Iterable[str]):
        """ Reset the clocks for a set of axes """
        self._moved_at.update({ax: None for ax in axis_iter})


def string_to_hex(val: str, min_length: int = 0) -> str:
    """
    Create a hex representation of val. The end of the result will be padded
    with "0" until min_length is reached.

    Args:
        val: The string to convert.
        min_length: The minimum length of result. "0" will be used as
            padding. Default is no minimum length and no padding.

    Returns:
        Hex string
    """
    hex_string = binascii.hexlify(val.encode()).decode()
    hex_string_length = len(hex_string)
    if hex_string_length < min_length:
        return hex_string + "0" * (min_length - hex_string_length)
    return hex_string



"""The interface that implements ProtocolContext."""

from __future__ import annotations

from abc import abstractmethod, ABC
from dataclasses import dataclass
from typing import (Dict, Optional)

from opentrons import types
from opentrons.hardware_control import API, SynchronousAdapter
from opentrons.protocols.geometry.deck import Deck
from opentrons.protocols.geometry.deck_item import DeckItem
from opentrons.protocols.geometry.module_geometry import (
    ModuleGeometry, ModuleType)
from opentrons.protocols.context.instrument \
    import AbstractInstrument
from opentrons.protocols.api_support.util import (
    AxisMaxSpeeds, HardwareManager)
from opentrons.protocols.context.labware import \
    AbstractLabware
from opentrons_shared_data.labware.dev_types import LabwareDefinition


InstrumentDict = Dict[types.Mount, Optional[AbstractInstrument]]


@dataclass(frozen=True)
class LoadModuleResult:
    """The result of load_module"""
    type: ModuleType
    geometry: ModuleGeometry
    module: SynchronousAdapter


class AbstractProtocol(ABC):

    @abstractmethod
    def get_bundled_data(self) -> Dict[str, bytes]:
        """Get a mapping of name to contents"""
        ...

    @abstractmethod
    def get_bundled_labware(self) -> Optional[Dict[str, LabwareDefinition]]:
        ...

    @abstractmethod
    def get_extra_labware(self) -> Optional[Dict[str, LabwareDefinition]]:
        ...

    @abstractmethod
    def cleanup(self) -> None:
        ...

    @abstractmethod
    def get_max_speeds(self) -> AxisMaxSpeeds:
        ...

    @abstractmethod
    def get_hardware(self) -> HardwareManager:
        ...

    @abstractmethod
    def connect(self, hardware: API) -> None:
        ...

    @abstractmethod
    def disconnect(self) -> None:
        ...

    @abstractmethod
    def is_simulating(self) -> bool:
        ...

    @abstractmethod
    def load_labware_from_definition(
            self,
            labware_def: LabwareDefinition,
            location: types.DeckLocation,
            label: Optional[str],
    ) -> AbstractLabware:
        ...

    @abstractmethod
    def load_labware(
            self,
            load_name: str,
            location: types.DeckLocation,
            label: Optional[str],
            namespace: Optional[str],
            version: Optional[int],
    ) -> AbstractLabware:
        ...

    @abstractmethod
    def load_module(
            self,
            module_name: str,
            location: Optional[types.DeckLocation],
            configuration: Optional[str]) -> Optional[LoadModuleResult]:
        ...

    @abstractmethod
    def get_loaded_modules(self) -> Dict[int, LoadModuleResult]:
        ...

    @abstractmethod
    def load_instrument(
            self,
            instrument_name: str,
            mount: types.Mount,
            replace: bool) -> AbstractInstrument:
        ...

    @abstractmethod
    def get_loaded_instruments(self) -> InstrumentDict:
        ...

    @abstractmethod
    def pause(self,
              msg: Optional[str]) -> None:
        ...

    @abstractmethod
    def resume(self) -> None:
        ...

    @abstractmethod
    def comment(self,
                msg: str) -> None:
        ...

    @abstractmethod
    def delay(self,
              seconds: float,
              msg: Optional[str]) -> None:
        ...

    @abstractmethod
    def home(self) -> None:
        ...

    @abstractmethod
    def get_deck(self) -> Deck:
        ...

    @abstractmethod
    def get_fixed_trash(self) -> DeckItem:
        ...

    @abstractmethod
    def set_rail_lights(self, on: bool) -> None:
        ...

    @abstractmethod
    def get_rail_lights_on(self) -> bool:
        ...

    @abstractmethod
    def door_closed(self) -> bool:
        ...

    @abstractmethod
    def get_last_location(
        self,
        mount: Optional[types.Mount] = None,
    ) -> Optional[types.Location]:
        ...

    @abstractmethod
    def set_last_location(
        self,
        location: Optional[types.Location],
        mount: Optional[types.Mount] = None,
    ) -> None:
        ...



"""An experimental, in-progress rewrite of the public Python Protocol API v2.

This rewrite is to implement the API in terms of the newer Protocol Engine
back-end architecture.

Unless documented otherwise, the behavior of everything should be the same as
APIv2. There may be variation in behavior that was underspecified in APIv2--
we'll try to call this out when we know about it, but there are no guarantees
that those callouts are comprehensive.

This code is totally unsupported. To do science on a robot, use the stable
`opentrons.protocol_api` instead.
"""

# todo(mm, 2021-04-09): The APIv2 analogue to this file exposes some, but not
# all, package members at the top level. For example, you can access
# protocol_api.InstrumentContext, but not protocol_api.Point. We should take
# advantage of the APIv3 break to consolidate the API namespaces.

from .protocol_context import ProtocolContext
from .pipette_context import PipetteContext
from .instrument_context import InstrumentContext
from .labware import Labware
from .well import Well
from .errors import InvalidPipetteNameError, InvalidMountError
from .types import (
    DeckSlotName,
    LabwareParameters,
    PipetteName,
    Point,
    Mount,
    DeprecatedMount,
)

__all__ = [
    # Protocol API classes
    "ProtocolContext",
    "PipetteContext",
    "InstrumentContext",
    "Labware",
    "Well",
    # Protocol API errors
    "InvalidPipetteNameError",
    "InvalidMountError",
    # Protocol API types and data classes
    "DeckSlotName",
    "LabwareParameters",
    "PipetteName",
    "Point",
    "Mount",
    "DeprecatedMount",
]



from typing import cast, List, Optional, Tuple
from opentrons.types import Point
from opentrons.protocol_api import (
    labware, InstrumentContext, ProtocolContext)

from opentrons.protocols.geometry import module_geometry


def _get_parent_slot_and_position(
        labware_obj: labware.Labware) -> Tuple[str, Optional[Point]]:
    if isinstance(labware_obj.parent, (module_geometry.ModuleGeometry)):
        return (
            cast(str, labware_obj.parent.parent),
            labware_obj.parent.labware_offset)
    else:
        return (cast(str, labware_obj.parent), None)


class Container:
    def __init__(
            self,
            container: labware.Labware,
            instruments: List[InstrumentContext],
            context: ProtocolContext) -> None:
        self._container = container
        self._context = context
        self.id = id(container)
        self.labware_offset_from_slot = 0

        # will be labware's load name or label
        self.name = container.name
        # type must be load_name so client can load correct definition
        self.type = container.load_name
        slot, position = _get_parent_slot_and_position(container)
        self.slot = slot
        self.position = position
        self.is_legacy = False
        self.is_tiprack = container.is_tiprack
        self.definition_hash = labware.get_labware_hash_with_parent(
            container)
        self.instruments = [
            Instrument(instrument, [], self._context)
            for instrument in instruments]


class Instrument:
    def __init__(
            self,
            instrument: InstrumentContext,
            containers: List[labware.Labware],
            context: ProtocolContext) -> None:
        containers = containers or []
        self._instrument = instrument
        self._context = context

        self.id = id(instrument)
        # The name element here is actually the pipette model for historical
        # reasons
        self.name = instrument.model
        self.model_name = instrument.name
        self.channels = instrument.channels
        self.mount = instrument.mount
        self.containers = [
            Container(container, [], self._context)
            for container in containers
        ]
        self.tip_racks = [
            Container(container, [], self._context)
            for container in instrument.tip_racks]
        if context:
            self.tip_racks.extend([
                c for c in self.containers if c._container.is_tiprack])
        self.requested_as = instrument.requested_as


class Module:
    def __init__(
            self,
            module: module_geometry.ModuleGeometry,
            context: ProtocolContext) -> None:
        self.id = id(module)
        _type_lookup = {
            module_geometry.ModuleType.MAGNETIC: 'magdeck',
            module_geometry.ModuleType.TEMPERATURE: 'tempdeck',
            module_geometry.ModuleType.THERMOCYCLER: 'thermocycler'}
        self.name = _type_lookup[module.module_type]
        self.model = module.model.value
        self.slot = module.parent
        self._context = context



"""ProtocolEngine class definition."""
from __future__ import annotations
from typing import Union

from opentrons.hardware_control.api import API as HardwareAPI
from opentrons.util.helpers import utc_now

from .errors import ProtocolEngineError, UnexpectedProtocolError
from .execution import CommandHandlers
from .resources import ResourceProviders
from .state import StateStore, StateView
from .commands import (
    CommandRequestType,
    CompletedCommandType,
    FailedCommandType,
)


class ProtocolEngine:
    """Main ProtocolEngine class.

    A ProtocolEngine instance holds the state of a protocol as it executes,
    and manages calls to a command executor that actually implements the logic
    of the commands themselves.
    """

    state_store: StateStore
    _handlers: CommandHandlers

    @classmethod
    async def create(cls, hardware: HardwareAPI) -> ProtocolEngine:
        """Create a ProtocolEngine instance."""
        resources = ResourceProviders.create()

        # TODO(mc, 2020-11-18): check short trash FF
        # TODO(mc, 2020-11-18): consider moving into a StateStore.create factory
        deck_def = await resources.deck_data.get_deck_definition()
        fixed_labware = await resources.deck_data.get_deck_fixed_labware(deck_def)

        state_store = StateStore(
            deck_definition=deck_def,
            deck_fixed_labware=fixed_labware
        )

        handlers = CommandHandlers.create(
            resources=resources,
            hardware=hardware,
            state=StateView.create_view(state_store),
        )

        return cls(state_store=state_store, handlers=handlers)

    def __init__(
        self,
        state_store: StateStore,
        handlers: CommandHandlers,
    ) -> None:
        """Initialize a ProtocolEngine instance.

        This constructor does not inject provider implementations. Prefer the
        ProtocolEngine.create factory classmethod.
        """
        self.state_store = state_store
        self._handlers = handlers

    async def execute_command(
        self,
        request: CommandRequestType,
        command_id: str,
    ) -> Union[CompletedCommandType, FailedCommandType]:
        """Execute a command request, waiting for it to complete."""
        cmd_impl = request.get_implementation()
        created_at = utc_now()
        cmd = cmd_impl.create_command(created_at).to_running(created_at)
        done_cmd: Union[CompletedCommandType, FailedCommandType]

        # store the command prior to execution
        self.state_store.handle_command(cmd, command_id=command_id)

        # execute the command
        try:
            result = await cmd_impl.execute(self._handlers)
            completed_at = utc_now()
            done_cmd = cmd.to_completed(result, completed_at)

        except Exception as error:
            failed_at = utc_now()
            if not isinstance(error, ProtocolEngineError):
                error = UnexpectedProtocolError(error)
            done_cmd = cmd.to_failed(error, failed_at)

        # store the done command
        self.state_store.handle_command(done_cmd, command_id=command_id)

        return done_cmd

    def add_command(self, request: CommandRequestType) -> None:
        """Add a command to ProtocolEngine."""
        # TODO(spp, 2020-05-13):
        #   Generate a UUID to be used as command_id for each command added.
        raise NotImplementedError



"""Command execution module."""

from .command_handlers import CommandHandlers
from .equipment import EquipmentHandler, LoadedLabware, LoadedPipette
from .movement import MovementHandler
from .pipetting import PipettingHandler

__all__ = [
    "CommandHandlers",
    "EquipmentHandler",
    "LoadedLabware",
    "LoadedPipette",
    "MovementHandler",
    "PipettingHandler",
]



import json
import logging
import os
import sys
from functools import lru_cache
from typing import Any, Dict, Mapping, Tuple, Union, \
    Optional, TYPE_CHECKING, NamedTuple

from opentrons.config import CONFIG, ARCHITECTURE, SystemArchitecture
from opentrons.system import log_control

if TYPE_CHECKING:
    from pathlib import Path

log = logging.getLogger(__name__)


SettingsMap = Dict[str, Optional[bool]]


class SettingException(Exception):
    def __init__(self, message, error):
        super(Exception, self).__init__(message)
        self.error = error


class SettingsData(NamedTuple):
    settings_map: SettingsMap
    version: int


class SettingDefinition:
    def __init__(self, _id: str, title: str, description: str,
                 old_id: str = None,
                 restart_required: bool = False,
                 show_if: Tuple[str, bool] = None):
        self.id = _id
        #: The id of the setting for programmatic access through
        #: get_adv_setting
        self.old_id = old_id
        #: the old id before migration, if any
        self.title = title
        #: User facing title
        self.description = description
        #: User facing description
        self.restart_required = restart_required
        #: True if the user must restart
        self.show_if = show_if
        #: A tuple of (other setting id, setting value) that must match reality
        #: to show this setting in http endpoints

    def __repr__(self):
        return '{}: {}'.format(self.__class__, self.id)

    def should_show(self) -> bool:
        """
        Use show_if attribute to determine if setting should be presented
        to users
        """
        if not self.show_if:
            return True
        return get_setting_with_env_overload(self.show_if[0]) == \
            self.show_if[1]

    async def on_change(self, value: Optional[bool]):
        """
        An opportunity for side effects as a result of change a setting
        value
        """
        if self.restart_required:
            set_restart_required()


class DisableLogIntegrationSettingDefinition(SettingDefinition):
    def __init__(self):
        super().__init__(
            _id='disableLogAggregation',
            title='Disable Opentrons Log Collection',
            description='Prevent the robot from sending its logs to Opentrons'
                        ' for analysis. Opentrons uses these logs to'
                        ' troubleshoot robot issues and spot error trends.')

    async def on_change(self, value: Optional[bool]):
        """Special side effect for this setting"""
        if ARCHITECTURE == SystemArchitecture.BUILDROOT:
            code, stdout, stderr = await log_control.set_syslog_level(
                'emerg' if value else 'info'
            )
            if code != 0:
                log.error(
                    f"Could not set log control: {code}: stdout={stdout}"
                    f" stderr={stderr}")
                raise SettingException(
                    f'Failed to set log upstreaming: {code}',
                    'log-config-failure'
                )
        await super().on_change(value)


class Setting(NamedTuple):
    value: Optional[bool]
    definition: SettingDefinition


# If you add or remove any settings here BE SURE TO ADD A MIGRATION below.
# You will also need to update the migration tests in:
# api/tests/opentrons/config/test_advanced_settings_migration.py
settings = [
    SettingDefinition(
        _id='shortFixedTrash',
        old_id='short-fixed-trash',
        title='Short (55mm) fixed trash',
        description='Trash box is 55mm tall (rather than the 77mm default)'
    ),
    SettingDefinition(
        _id='calibrateToBottom',
        old_id='calibrate-to-bottom',
        title='Calibrate to bottom',
        description='Calibrate using the bottom-center of well A1 for each'
                    ' labware (rather than the top-center)'
    ),
    SettingDefinition(
        _id='deckCalibrationDots',
        old_id='dots-deck-type',
        title='Deck calibration to dots',
        description='Perform deck calibration to dots rather than crosses, for'
                    ' robots that do not have crosses etched on the deck'
    ),
    SettingDefinition(
        _id='useProtocolApi2',
        title='Use Protocol API version 2',
        description='Deprecated feature flag'
    ),
    SettingDefinition(
        _id='disableHomeOnBoot',
        old_id='disable-home-on-boot',
        title='Disable home on boot',
        description='Prevent robot from homing motors on boot'
    ),
    SettingDefinition(
        _id='useOldAspirationFunctions',
        title='Use older aspirate behavior',
        description='Aspirate with the less accurate volumetric calibrations'
                    ' that were used before version 3.7.0. Use this if you'
                    ' need consistency with pre-v3.7.0 results. This only'
                    ' affects GEN1 P10S, P10M, P50S, P50M, and P300S pipettes.'
    ),
    SettingDefinition(
        _id='enableDoorSafetySwitch',
        title='Enable robot door safety switch',
        description="Automatically pause protocols when robot door opens. "
                    "Opening the robot door during a run will "
                    "pause your robot only after it has completed its "
                    "current motion."
    ),
    SettingDefinition(
        _id='enableTipLengthCalibration',
        title='Enable Under-Development Calibration Flows',
        description='Do not activate this unless you are a developer. '
                    'Enables the in-progress robot calibration flows '
                    'for tip length, deck, and instrument calibration.'
    ),
    SettingDefinition(
        _id='enableHttpProtocolSessions',
        title='Enable Experimental HTTP Protocol Sessions',
        description='Do not activate this unless you are a developer. '
                    'Activating this will disable protocol running from the '
                    'Opentrons application.',
        restart_required=True,
    ),
    SettingDefinition(
        _id='enableFastProtocolUpload',
        title='Enable Experimental Fast Protocol Upload',
        description='Enabling this flag will skip simulation for a faster '
                    'upload. The protocol will be analyzed for syntax errors, '
                    'run steps, and equipment requirements. This feature '
                    'should only be used if a protocol has been simulated '
                    'before.',
        restart_required=False,
    ),
    SettingDefinition(
        _id="enableProtocolEngine",
        title="Enable Experimental Protocol Engine",
        description=(
            "Do not enable. This is an Opentrons-internal setting to test "
            "new protocol execution logic. This feature is not yet complete; "
            "your protocols will break if you enable this setting."
        ),
        restart_required=False,
    )
]

if ARCHITECTURE == SystemArchitecture.BUILDROOT:
    settings.append(DisableLogIntegrationSettingDefinition())


settings_by_id: Dict[str, SettingDefinition] = \
    {s.id: s for s in settings}
settings_by_old_id: Dict[str, SettingDefinition] = \
    {s.old_id: s for s in settings if s.old_id}


def get_adv_setting(setting: str) -> Optional[Setting]:
    setting = _clean_id(setting)
    s = get_all_adv_settings()
    return s.get(setting, None)


@lru_cache(maxsize=1)
def get_all_adv_settings() -> Dict[str, Setting]:
    """Get all the advanced setting values and definitions"""
    settings_file = CONFIG['feature_flags_file']

    values, _ = _read_settings_file(settings_file)

    return {
        key: Setting(value=value, definition=settings_by_id[key])
        for key, value in values.items() if key in settings_by_id
    }


async def set_adv_setting(_id: str, value: Optional[bool]):
    _id = _clean_id(_id)
    settings_file = CONFIG['feature_flags_file']
    setting_data = _read_settings_file(settings_file)
    if _id not in setting_data.settings_map:
        raise ValueError(f"{_id} is not recognized")
    # Side effecting
    await settings_by_id[_id].on_change(value)

    setting_data.settings_map[_id] = value
    _write_settings_file(setting_data.settings_map,
                         setting_data.version,
                         settings_file)
    # Clear the lru cache
    get_all_adv_settings.cache_clear()


def _clean_id(_id: str) -> str:
    if _id in settings_by_old_id.keys():
        _id = settings_by_old_id[_id].id
    return _id


def _read_json_file(path: Union[str, 'Path']) -> Dict[str, Any]:
    try:
        with open(path, 'r') as fd:
            data = json.load(fd)
    except FileNotFoundError:
        data = {}
    except json.JSONDecodeError as e:
        sys.stderr.write(
            f'Could not load advanced settings file {path}: {e}\n')
        data = {}
    return data


def _read_settings_file(settings_file: 'Path') -> SettingsData:
    """
    Read the settings file, which is a json object with settings IDs as keys
    and boolean values. For each key, look up the `Settings` object with that
    key. If the key is one of the old IDs (kebab case), replace it with the
    new ID and rewrite the settings file

    :param settings_file: the path to the settings file
    :return: a dict with all new settings IDs as the keys, and boolean values
        (the values stored in the settings file, or `False` if the key was not
        found). Along with the version.
    """
    # Read settings from persistent file
    data = _read_json_file(settings_file)
    settings, version = _migrate(data)
    settings = _ensure(settings)

    if data.get('_version') != version:
        _write_settings_file(settings, version, settings_file)

    return SettingsData(settings_map=settings, version=version)


def _write_settings_file(data: Mapping[str, Any],
                         version: int,
                         settings_file: 'Path'):
    try:
        with settings_file.open('w') as fd:
            json.dump({**data, '_version': version}, fd)
            fd.flush()
            os.fsync(fd.fileno())
    except OSError:
        log.exception(
            f'Failed to write advanced settings file to {settings_file}')


def _migrate0to1(previous: Mapping[str, Any]) -> SettingsMap:
    """
    Migrate to version 1 of the feature flags file. Replaces old IDs with new
    IDs and sets any False values to None
    """
    next: SettingsMap = {}

    for s in settings:
        id = s.id
        old_id = s.old_id

        if previous.get(id) is True:
            next[id] = True
        elif old_id and previous.get(old_id) is True:
            next[id] = True
        else:
            next[id] = None

    return next


def _migrate1to2(previous: SettingsMap) -> SettingsMap:
    """
    Migration to version 2 of the feature flags file. Adds the
    disableLogAggregation config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap['disableLogAggregation'] = None
    return newmap


def _migrate2to3(previous: SettingsMap) -> SettingsMap:
    """
    Migration to version 3 of the feature flags file. Adds the
    enableApi1BackCompat config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap['enableApi1BackCompat'] = None
    newmap['useProtocolApi2'] = None
    return newmap


def _migrate3to4(previous: SettingsMap) -> SettingsMap:
    """
    Migration to version 4 of the feature flags file. Adds the
    useV1HttpApi config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap['useV1HttpApi'] = None
    return newmap


def _migrate4to5(previous: SettingsMap) -> SettingsMap:
    """
    Migration to version 5 of the feature flags file. Adds the
    enableDoorSafetyFeature config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap['enableDoorSafetySwitch'] = None
    return newmap


def _migrate5to6(previous: SettingsMap) -> SettingsMap:
    """
    Migration to version 6 of the feature flags file. Adds the
    enableTipLengthCalibration config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap['enableTipLengthCalibration'] = None
    return newmap


def _migrate6to7(previous: SettingsMap) -> SettingsMap:
    """
    Migration to version 7 of the feature flags file. Adds the
    enableHttpProtocolSessions config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap['enableHttpProtocolSessions'] = None
    return newmap


def _migrate7to8(previous: SettingsMap) -> SettingsMap:
    """
    Migration to version 8 of the feature flags file. Adds the
    enableFastProtocolUpload config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap['enableFastProtocolUpload'] = None
    return newmap


def _migrate8to9(previous: SettingsMap) -> SettingsMap:
    """
    Migration to version 8 of the feature flags file. Adds the
    enableFastProtocolUpload config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["enableProtocolEngine"] = None
    return newmap


_MIGRATIONS = [_migrate0to1, _migrate1to2, _migrate2to3, _migrate3to4,
               _migrate4to5, _migrate5to6, _migrate6to7, _migrate7to8,
               _migrate8to9]
"""
List of all migrations to apply, indexed by (version - 1). See _migrate below
for how the migration functions are applied. Each migration function should
return a new dictionary (rather than modify their input)
"""


def _migrate(data: Mapping[str, Any]) -> SettingsData:
    """
    Check the version integer of the JSON file data a run any necessary
    migrations to get us to the latest file format. Returns dictionary of
    settings and version migrated to
    """
    next = dict(data)
    version = next.pop('_version', 0)
    target_version = len(_MIGRATIONS)
    migrations = _MIGRATIONS[version:]

    if len(migrations) > 0:
        log.info(
            "Migrating advanced settings from version {} to {}"
            .format(version, target_version))

    for m in migrations:
        next = m(next)

    return SettingsData(settings_map=next, version=target_version)


def _ensure(data: Mapping[str, Any]) -> SettingsMap:
    """
    Even after migration, we may have an invalid file. For instance,
    we may have _downgraded_. Make sure all required keys are present.
    """
    newdata = {k: v for k, v in data.items()}
    for s in settings:
        if s.id not in newdata:
            newdata[s.id] = None
    return newdata


def get_setting_with_env_overload(setting_name) -> bool:
    env_name = 'OT_API_FF_' + setting_name
    if env_name in os.environ:
        return os.environ[env_name].lower() in {'1', 'true', 'on'}
    else:
        s = get_adv_setting(setting_name)
        return s.value is True if s is not None else False


_SETTINGS_RESTART_REQUIRED = False
# This is a bit of global state that indicates whether a setting has changed
# that requires a restart. It's OK for this to be global because the behavior
# it's catching is global - changing the kind of setting that requires a
# a restart anywhere, even if you theoretically have two servers running in
# the same process, will require _all_ of them to be restarted.


def is_restart_required() -> bool:
    return _SETTINGS_RESTART_REQUIRED


def set_restart_required() -> bool:
    """Set the restart required flag."""
    global _SETTINGS_RESTART_REQUIRED
    _SETTINGS_RESTART_REQUIRED = True
    return _SETTINGS_RESTART_REQUIRED



from __future__ import annotations
from typing import TYPE_CHECKING, List, Union, overload


from .helpers import stringify_location, listify
from . import types as command_types

from opentrons.types import Location
from opentrons.protocols.api_support.util import FlowRates

if TYPE_CHECKING:
    from opentrons.protocol_api import InstrumentContext
    from opentrons.protocol_api.labware import Well


def home(mount: str) -> command_types.HomeCommand:
    text = f'Homing pipette plunger on mount {mount}'
    return {
        'name': command_types.HOME,
        'payload': {
            'axis': mount,
            'text': text
        }
    }


def aspirate(
        instrument: InstrumentContext,
        volume: float,
        location: Union[Well, Location],
        rate: float) -> command_types.AspirateCommand:
    location_text = stringify_location(location)
    template = 'Aspirating {volume} uL from {location} at {flow} uL/sec'
    flow_rate = rate * FlowRates(instrument._implementation).aspirate
    text = template.format(
        volume=float(volume), location=location_text, flow=flow_rate)

    return {
        'name': command_types.ASPIRATE,
        'payload': {
            'instrument': instrument,
            'volume': volume,
            'location': location,
            'rate': rate,
            'text': text
        }
    }


def dispense(
        instrument: InstrumentContext,
        volume: float,
        location: Union[Well, Location],
        rate: float) -> command_types.DispenseCommand:
    location_text = stringify_location(location)
    template = 'Dispensing {volume} uL into {location} at {flow} uL/sec'
    flow_rate = rate * FlowRates(instrument._implementation).dispense
    text = template.format(
        volume=float(volume), location=location_text, flow=flow_rate)

    return {
        'name': command_types.DISPENSE,
        'payload': {
            'instrument': instrument,
            'volume': volume,
            'location': location,
            'rate': rate,
            'text': text
        }
    }


def consolidate(
        instrument: InstrumentContext,
        volume: Union[float, List[float]],
        source: List[Union[Location, Well]],
        dest: Union[Location, Well]) -> command_types.ConsolidateCommand:
    text = 'Consolidating {volume} from {source} to {dest}'.format(
        volume=transform_volumes(volume),
        source=stringify_location(source),
        dest=stringify_location(dest)
    )
    locations: List[Union[Location, Well]] = listify(source) + listify(dest)
    return {
        'name': command_types.CONSOLIDATE,
        'payload': {
            'instrument': instrument,
            'locations': locations,
            'volume': volume,
            'source': source,
            'dest': dest,
            'text': text
        }
    }


def distribute(
        instrument: InstrumentContext,
        volume: Union[float, List[float]],
        source: Union[Location, Well],
        dest: List[Union[Location, Well]]) -> command_types.DistributeCommand:
    text = 'Distributing {volume} from {source} to {dest}'.format(
        volume=transform_volumes(volume),
        source=stringify_location(source),
        dest=stringify_location(dest)
    )
    locations: List[Union[Location, Well]] = listify(source) + listify(dest)
    return {
        'name': command_types.DISTRIBUTE,
        'payload': {
            'instrument': instrument,
            'locations': locations,
            'volume': volume,
            'source': source,
            'dest': dest,
            'text': text
        }
    }


def transfer(
        instrument: InstrumentContext,
        volume: Union[float, List[float]],
        source: List[Union[Location, Well]],
        dest: List[Union[Location, Well]]) -> command_types.TransferCommand:
    text = 'Transferring {volume} from {source} to {dest}'.format(
        volume=transform_volumes(volume),
        source=stringify_location(source),
        dest=stringify_location(dest)
    )
    locations: List[Union[Location, Well]] = listify(source) + listify(dest)
    return {
        'name': command_types.TRANSFER,
        'payload': {
            'instrument': instrument,
            'locations': locations,
            'volume': volume,
            'source': source,
            'dest': dest,
            'text': text
        }
    }


@overload
def transform_volumes(volumes: Union[float, int]) -> float: ...


@overload
def transform_volumes(volumes: List[Union[float]]) -> List[float]: ...


def transform_volumes(volumes):
    if not isinstance(volumes, list):
        return float(volumes)
    else:
        return [float(vol) for vol in volumes]


def mix(instrument: InstrumentContext,
        repetitions: int,
        volume: float,
        location: Union[Well, Location, None]) -> command_types.MixCommand:
    text = 'Mixing {repetitions} times with a volume of {volume} ul'.format(
        repetitions=repetitions, volume=float(volume)
    )
    return {
        'name': command_types.MIX,
        'payload': {
            'instrument': instrument,
            'location': location,
            'volume': volume,
            'repetitions': repetitions,
            'text': text
        }
    }


def blow_out(
        instrument: InstrumentContext,
        location: Union[Well, Location, None]) -> command_types.BlowOutCommand:
    location_text = stringify_location(location)
    text = 'Blowing out'

    if location is not None:
        text += ' at {location}'.format(location=location_text)

    return {
        'name': command_types.BLOW_OUT,
        'payload': {
            'instrument': instrument,
            'location': location,
            'text': text
        }
    }


def touch_tip(instrument: InstrumentContext) -> command_types.TouchTipCommand:
    text = 'Touching tip'

    return {
        'name': command_types.TOUCH_TIP,
        'payload': {
            'instrument': instrument,
            'text': text
        }
    }


def air_gap() -> command_types.AirGapCommand:
    text = 'Air gap'
    return {
        'name': command_types.AIR_GAP,
        'payload': {
            'text': text
        }
    }


def return_tip() -> command_types.ReturnTipCommand:
    text = 'Returning tip'
    return {
        'name': command_types.RETURN_TIP,
        'payload': {
            'text': text
        }
    }


def pick_up_tip(
        instrument: InstrumentContext,
        location: Union[Location, Well]) -> command_types.PickUpTipCommand:
    location_text = stringify_location(location)
    text = f'Picking up tip from {location_text}'
    return {
        'name': command_types.PICK_UP_TIP,
        'payload': {
            'instrument': instrument,
            'location': location,
            'text': text
        }
    }


def drop_tip(
        instrument: InstrumentContext,
        location: Union[Location, Well]) -> command_types.DropTipCommand:
    location_text = stringify_location(location)
    text = 'Dropping tip into {location}'.format(location=location_text)
    return {
        'name': command_types.DROP_TIP,
        'payload': {
            'instrument': instrument,
            'location': location,
            'text': text
        }
    }


def move_to(
        instrument: InstrumentContext,
        location: Union[Location, Well]) -> command_types.MoveToCommand:
    location_text = stringify_location(location)
    text = 'Moving to {location}'.format(location=location_text)
    return {
        'name': command_types.MOVE_TO,
        'payload': {
            'instrument': instrument,
            'location': location,
            'text': text
        }
    }



from typing import Optional, Generic, TypeVar

TEMPERATURE_ROOM = 23


ValueType = TypeVar('ValueType')


class OptionalValue(Generic[ValueType]):
    """
    A class that serializes optional values.

    Modules represent a null value as 'none'. For example this response from
    the thermocycler means the hold time is not set:
        H:none T:1.23
    """
    _value: Optional[ValueType]

    def __init__(self, value: Optional[ValueType] = None):
        self._value = value

    @property
    def val(self) -> Optional[ValueType]:
        return self._value

    @val.setter
    def val(self, value: Optional[ValueType]) -> None:
        self._value = value

    def __repr__(self) -> str:
        return "none" if self._value is None else str(self._value)

    def __eq__(self, other) -> bool:
        if not isinstance(other, OptionalValue):
            return False

        return other.val == self.val
